import unittest
from octopus_python_client import hello_world


class Test(unittest.TestCase):

    def setUp(self):
        pass

    def test_text_compare_true(self):
        self.assertTrue(hello_world.text_compare("asdf", "asdf"))

    def test_text_compare_false(self):
        self.assertFalse(hello_world.text_compare("asdf", "foo"))


if __name__ == '__main__':
    unittest.main()



import logging

from octopus_python_client.common import item_type_projects, get_one_type_save, get_single_item_by_name_or_id_save, \
    delete_single_item_by_name_or_id, update_single_item_save, create_single_item_from_local_file, \
    clone_single_item_from_remote_item, item_type_library_variable_sets, included_library_variable_set_ids_key, \
    id_key, name_key, put_single_item_save, get_one_type_ignore_error, get_list_items_from_all_items, log_info_print
from octopus_python_client.processes import clone_deployment_process
from octopus_python_client.utilities.helper import find_item, compare_lists

logger = logging.getLogger(__name__)


def get_all_projects(space_id=None):
    return get_one_type_save(item_type=item_type_projects, space_id=space_id)


def get_project(project_literal_name=None, space_id=None):
    log_info_print(local_logger=logger, msg=f"get and save project {project_literal_name} in space {space_id}...")
    project = get_single_item_by_name_or_id_save(item_type=item_type_projects, item_name=project_literal_name,
                                                 space_id=space_id)
    return project


def update_project(project_literal_name=None, space_id=None):
    update_single_item_save(item_type=item_type_projects, item_name=project_literal_name, space_id=space_id)


def create_project_from_local_file(project_literal_name=None, local_project_name=None, space_id=None):
    return create_single_item_from_local_file(item_type=item_type_projects, item_name=project_literal_name,
                                              local_item_name=local_project_name, space_id=space_id)


def clone_project(project_literal_name=None, base_project_name=None, space_id=None):
    log_info_print(local_logger=logger,
                   msg=f"clone project from {base_project_name} to {project_literal_name} inside space {space_id}")
    clone_single_item_from_remote_item(item_type=item_type_projects, item_name=project_literal_name,
                                       base_item_name=base_project_name, space_id=space_id)
    clone_deployment_process(project_literal_name=project_literal_name, base_project_name=base_project_name,
                             space_id=space_id)


def delete_project(project_literal_name=None, space_id=None):
    log_info_print(local_logger=logger, msg=f"delete project {project_literal_name} in space {space_id}")
    delete_single_item_by_name_or_id(item_type=item_type_projects, item_name=project_literal_name, space_id=space_id)


def process_suffix(name=None, suffix=None, remove_suffix=False):
    dot_suffix = '.' + suffix
    if not remove_suffix:
        return name + dot_suffix
    elif name.endswith(dot_suffix):
        return name[:-len(dot_suffix)]
    return name


def project_update_variable_sets(project_literal_name=None, suffix=None, space_id=None, remove_suffix=False):
    if not project_literal_name or not suffix:
        raise ValueError("project name or suffix must not be empty")
    log_info_print(local_logger=logger, msg=f"updating {space_id}'s project {project_literal_name}'s variable sets by "
                                            f"{'removing' if remove_suffix else 'adding'} a suffix {suffix}...")
    all_variable_sets = get_one_type_ignore_error(item_type=item_type_library_variable_sets, space_id=space_id)
    list_variable_sets = get_list_items_from_all_items(all_items=all_variable_sets)
    project = get_project(project_literal_name, space_id=space_id)
    variable_sets_ids = project.get(included_library_variable_set_ids_key, [])
    logger.info("original variable sets id:")
    logger.info(variable_sets_ids)
    mapped_ids = []
    for variable_set_id in variable_sets_ids:
        variable_set = find_item(lst=list_variable_sets, key=id_key, value=variable_set_id)
        variable_set_name = variable_set.get(name_key)
        variable_set_name_suffix = process_suffix(name=variable_set_name, suffix=suffix, remove_suffix=remove_suffix)
        variable_set_suffix = find_item(lst=list_variable_sets, key=name_key, value=variable_set_name_suffix)
        if variable_set_suffix:
            logger.info(f"{variable_set_suffix} found in variable sets")
            mapped_ids.append(variable_set_suffix.get(id_key))
        else:
            logger.info(f"{variable_set_suffix} not found in variable sets")
            mapped_ids.append(variable_set_id)
    logger.info("mapped variable sets id:")
    logger.info(mapped_ids)
    no_change = compare_lists(variable_sets_ids, mapped_ids)
    if no_change:
        logger.info(f"The variable sets have no change")
        return project
    project[included_library_variable_set_ids_key] = mapped_ids
    return put_single_item_save(item_type=item_type_projects, payload=project, space_id=space_id)



import copy
import logging
import re
from time import gmtime, strftime

from octopus_python_client.common import name_key, tags_key, id_key, item_type_tag_sets, post_single_item_save, \
    item_type_projects, get_list_items_from_file, item_type_deployment_processes, deployment_process_id_key, config, \
    put_single_item_save, normal_cloneable_types, \
    get_list_items_from_all_items, version_key, must_have_types, item_type_library_variable_sets, item_type_variables, \
    variable_set_id_key, get_or_delete_single_item_by_id, item_type_tenant_variables, canonical_tag_name_key, \
    item_type_tags, tenant_id_key, item_type_migration, space_map, get_local_single_item_file, put_single_item, \
    item_type_tenants, slash_sign, underscore_sign, item_type_feeds, \
    secret_key_key, new_value_key, hyphen_sign, item_type_channels, project_id_key, \
    item_type_releases, item_type_artifacts, file_name_key, put_post_tenant_variables_save, \
    get_one_type_to_list, runbook_process_id_key, item_type_runbooks, \
    item_type_accounts, token_key, comma_sign, space_id_key, \
    published_runbook_snapshot_id_key, item_type_scoped_user_roles, user_role_id_key, team_id_key, \
    item_type_runbook_processes, runbook_process_prefix, item_id_prefix_to_type_dict, positive_integer_regex, \
    prepare_project_versioning_strategy, log_info_print, get_one_type_ignore_error, get_single_item_by_name_or_id
from octopus_python_client.utilities.helper import find_item, save_file, find_matched_sub_list, log_raise_value_error

logger = logging.getLogger(__name__)


class Migration:
    def __init__(self):
        self.__src_space_id = None
        self.__dst_space_id = None
        self.__dst_id_payload_dict = {}
        self.__dst_tenant_variables_payload_dict = {}
        self.__src_id_payload_dict = {}
        self.__src_id_type_dict = {}
        self.__src_id_vs_dst_id_dict = {}
        self.__src_tenant_variables_payload_dict = {}
        self.__type_dst_list_items_dict = {}
        self.__type_full_func_dict = {}
        self.__type_post_func_dict = {}
        self.__type_prep_func_dict = {}
        self.__type_src_list_items_dict = {}
        self.__item_types_set = set()

    # search the type in the space and see if the matched item already exists
    def __find_matched_dst_item_by_src_item(self, src_item_with_dst_ids=None, item_type=None):
        logger.info(f"Look for the matched destination {item_type} {src_item_with_dst_ids.get(name_key)} in "
                    f"{self.__dst_space_id}")
        dst_list_items = get_one_type_to_list(item_type=item_type, space_id=self.__dst_space_id)
        item_name = src_item_with_dst_ids.get(name_key)
        match_dict = None
        # "channels" and "runbooks" are special, the name is not unique across a space;
        # we must use both name and project id to find the match
        if item_type == item_type_channels or item_type == item_type_runbooks:
            match_dict = {name_key: item_name, project_id_key: src_item_with_dst_ids.get(project_id_key)}
        # type "releases" has no name and is unique by "Version" and "ProjectId"
        elif item_type == item_type_releases:
            match_dict = {version_key: src_item_with_dst_ids.get(version_key),
                          project_id_key: src_item_with_dst_ids.get(project_id_key)}
        # TODO cloning scopeduserroles is not working and may not be necessary; some space id is null;
        elif item_type == item_type_scoped_user_roles:
            match_dict = {user_role_id_key: src_item_with_dst_ids.get(user_role_id_key),
                          team_id_key: src_item_with_dst_ids.get(team_id_key)}
        # TODO type "artifacts" has no name and unique by Filename and ServerTaskId, so it is not cloneable
        elif item_type == item_type_artifacts:
            return find_item(lst=dst_list_items, key=file_name_key, value=src_item_with_dst_ids.get(file_name_key))
        elif item_name:
            match_dict = {name_key: item_name}
        else:
            log_raise_value_error(local_logger=logger, item=src_item_with_dst_ids,
                                  err=f"{item_type} does not have name or other keys for matching!")
        matched_sub_list = find_matched_sub_list(lst=dst_list_items, match_dict=match_dict, ignore_case=True)
        if matched_sub_list and len(matched_sub_list) > 1:
            log_raise_value_error(local_logger=logger, item=matched_sub_list,
                                  err=f"For {item_type} {item_name}, more than one item found in {self.__dst_space_id}")
        return matched_sub_list[0] if matched_sub_list else {}

    # For tagsets, if the destination space already have the same tagset,
    # each tag of each tagset must use the destination Tag ID not the source Tag ID,
    # so we have to replace each tag id with the destination ones,
    # otherwise it shows"Tag Id is in invalid format."
    # For additional Tag ID in source, we need to remove the ID from the source tag
    def __prepare_tag_set(self, src_item=None):
        src_list_tags = src_item.get(tags_key, [])
        # the source tag id should be all removed
        for src_tag in src_list_tags:
            src_tag.pop(id_key, None)
        dst_tag_set = self.__find_matched_dst_item_by_src_item(src_item_with_dst_ids=src_item,
                                                               item_type=item_type_tag_sets)
        if dst_tag_set:
            dst_list_tags = dst_tag_set.get(tags_key, [])
            for src_tag in src_list_tags:
                matched_dst_tag = find_item(lst=dst_list_tags, key=name_key, value=src_tag.get(name_key))
                if matched_dst_tag:
                    src_tag[id_key] = matched_dst_tag.get(id_key)

    # we do not want to clone the child items first;
    # child items will be created automatically when parent item is created
    def __prepare_project(self, src_item=None):
        logger.info(f"prepare {item_type_projects} {src_item.get(name_key)} for migrating to {self.__dst_space_id}")
        src_item.pop(deployment_process_id_key, None)
        src_item.pop(variable_set_id_key, None)
        prepare_project_versioning_strategy(project=src_item)

    # we do not want to clone the child items first;
    # child items will be created automatically when parent item is created
    def __prepare_library_variable_set(self, src_item=None):
        logger.info(f"prepare {item_type_library_variable_sets} {src_item.get(name_key)} for migrating to "
                    f"{self.__dst_space_id}")
        src_item.pop(variable_set_id_key, None)

    # SecretKey: NewValue: null; must be replaced by a placeholder string
    def __prepare_feed(self, src_item=None):
        logger.info(f"prepare {item_type_feeds} {src_item.get(name_key)} for migrating to {self.__dst_space_id}")
        if secret_key_key in src_item and not src_item.get(secret_key_key).get(new_value_key):
            src_item.get(secret_key_key)[new_value_key] = hyphen_sign
            logger.info(f"assigned a placeholder {hyphen_sign} to {new_value_key} for {secret_key_key}")

    # Token: NewValue: null; must be replaced by a placeholder string
    def __prepare_account(self, src_item=None):
        logger.info(f"prepare {item_type_accounts} {src_item.get(name_key)} for migrating to {self.__dst_space_id}")
        if token_key in src_item and not src_item.get(token_key).get(new_value_key):
            src_item.get(token_key)[new_value_key] = hyphen_sign
            logger.info(f"assigned a placeholder {hyphen_sign} to {new_value_key} for {token_key}")

    # runbook is a new type introduced in Octopus 2019.11; the older Octopus server may not support it
    # Need to upgrade Octopus server to the latest version
    def __prepare_runbook(self, src_item=None):
        logger.info(f"prepare {item_type_runbooks} {src_item.get(name_key)} {src_item.get(id_key)} for migrating to "
                    f"{self.__dst_space_id}: popping {runbook_process_id_key}")
        src_item.pop(runbook_process_id_key, None)
        src_item.pop(published_runbook_snapshot_id_key, None)

    def __clone_item_to_space(self, item_type=None, item_name=None, item_id=None):
        item_badge = item_name if item_name else item_id
        logger.info(f"clone {item_type} {item_badge} from {self.__src_space_id} to {self.__dst_space_id}")

        full_process = self.__type_full_func_dict.get(item_type)
        if full_process:
            return full_process(item_type=item_type, item_name=item_name, item_id=item_id)

        # find the source item in file/memory
        src_item = {}
        if item_name:
            src_list_items = self.__type_src_list_items_dict.get(item_type)
            src_item = find_item(lst=src_list_items, key=name_key, value=item_name)
        if not src_item:
            src_item = self.__src_id_payload_dict.get(item_id)
        if not src_item:
            raise ValueError(f"{item_type} {item_badge} does not exist in the source space")

        return self.__create_item_to_space(item_type=item_type, src_item=src_item)

    def __create_item_to_space(self, item_type=None, src_item=None):
        item_badge = src_item.get(name_key) if src_item.get(name_key) else src_item.get(id_key)
        log_info_print(local_logger=logger, msg=f"cloning {item_type} {item_badge} to space {self.__dst_space_id}")
        logger.info(f"preprocessing {item_type} {item_badge} with the new references in {self.__dst_space_id}...")
        # do not modify the items in memory
        src_item_copy = copy.deepcopy(src_item)

        # some type needs additional prep-processing
        prep_process = self.__type_prep_func_dict.get(item_type)
        if prep_process:
            logger.info(f"special preprocessing for {item_type} {item_badge} with function {prep_process}")
            prep_process(src_item=src_item_copy)

        # since the destination "Id: Self-2" has not been created and saved into map
        # we do not want to recursively replace the "Id: Self-1" of the payload;
        # it would cause infinite stack and overflow
        src_id_value = src_item_copy.pop(id_key, None)

        self.__replace_ids(dict_list=src_item_copy)

        logger.info(f"check if {item_type} {item_badge} already exists in {self.__dst_space_id}")
        dst_item = self.__find_matched_dst_item_by_src_item(src_item_with_dst_ids=src_item_copy, item_type=item_type)
        dst_item_exist = True if dst_item else False
        if dst_item_exist:
            if config.overwrite:
                logger.info(f"{self.__dst_space_id} already has {item_type} {item_badge}, overwriting it...")
                src_item_copy[id_key] = dst_item.get(id_key)

                # TODO bug in Octopus: PUT a runbook with null RunbookProcessId will remove RunbookProcessId from dst
                if item_type == item_type_runbooks and src_item.get(runbook_process_id_key):
                    matched_runbook_process_id = runbook_process_prefix + hyphen_sign + dst_item.get(id_key)
                    logger.info(f"Reassign process id {matched_runbook_process_id} for {src_item_copy.get(id_key)}")
                    src_item_copy[runbook_process_id_key] = matched_runbook_process_id

                dst_item = put_single_item_save(item_type=item_type, payload=src_item_copy,
                                                space_id=self.__dst_space_id)
            else:
                logger.info(f"{self.__dst_space_id} already has {item_type} {item_badge}, skipping it...")
        else:
            logger.info(f"{self.__dst_space_id} does not have {item_type} {item_badge}, so creating it...")
            dst_item = post_single_item_save(item_type=item_type, payload=src_item_copy, space_id=self.__dst_space_id)

        dst_id_value = dst_item.get(id_key)
        logger.info(f"add the id pair ({src_id_value}, {dst_id_value}) to the id map")
        self.__src_id_vs_dst_id_dict[src_id_value] = dst_id_value
        self.__dst_id_payload_dict[dst_id_value] = dst_item

        post_process = self.__type_post_func_dict.get(item_type)
        if post_process and (not dst_item_exist or config.overwrite):
            logger.info(f"Additional processing for {item_type} {item_badge} with function {post_process}")
            post_process(src_id=src_id_value, dst_id=dst_id_value)

        return dst_id_value

    def __clone_type_to_space(self, item_type=None):
        if not item_type:
            raise ValueError("item_type must not be empty!")
        log_info_print(local_logger=logger,
                       msg=f"---- cloning type {item_type} from {self.__src_space_id} to {self.__dst_space_id} -----")
        src_list_items = self.__type_src_list_items_dict.get(item_type)
        if not src_list_items:
            logger.warning(f"***** {item_type} has no items in {self.__src_space_id}, so skip processing it *****")
            return

        for src_item in src_list_items:
            self.__create_item_to_space(item_type=item_type, src_item=src_item)

        logger.info(f"Get the updated {item_type} in {self.__dst_space_id}")
        dst_all_items = get_one_type_ignore_error(item_type=item_type, space_id=self.__dst_space_id)
        dst_list_items = get_list_items_from_all_items(all_items=dst_all_items)
        self.__type_dst_list_items_dict[item_type] = dst_list_items

    # some id like "Environments-123" should be removed if this is not referenced anyway
    # most likely the entity has been removed and the reference is broken
    # TODO it is a dangerous operation; find a better way to warn people to fix the source space first
    def __check_broken_item_id(self, string=None):
        # if the string is one of the src or dst item ids, it is valid
        if string in self.__src_id_vs_dst_id_dict.keys() or string in self.__src_id_vs_dst_id_dict.values():
            return False
        for id_prefix, item_type in item_id_prefix_to_type_dict.items():
            pattern = r"{}".format("^" + id_prefix + positive_integer_regex)
            if re.match(pattern, string):
                logger.warning(f"Please remove the broken reference id {string} in source space {self.__src_space_id}")
                return True
        return False

    # recursively replace the old link id with the new link id
    # issue: if someone crazy names an environment name as "Environments-1", not "Development" or "Prod" etc,
    # it could cause the environment renamed as "Environment-10"
    # if id_links_map contains "Environments-1" as the key and "Environments-10" as the value
    def __replace_ids(self, dict_list=None):
        if isinstance(dict_list, dict):
            logger.info(f"dict_list is a dict, replace ids or delete broken ids")
            # directly use dict_list.items() or dict_list.keys() could cause unexpected result
            # due to dict.pop(key) while iterating
            keys = copy.deepcopy(list(dict_list.keys()))
            for key in keys:
                value = dict_list.get(key)
                if isinstance(value, str):
                    if value in self.__src_id_vs_dst_id_dict:
                        dict_list[key] = self.__src_id_vs_dst_id_dict.get(value)
                        logger.info(f"the reference value {value} has already been created in the destination space "
                                    f"and was replaced with {dict_list.get(key)}")
                    elif value in self.__src_id_payload_dict:
                        logger.info(f"the reference value {value} exists in the source space but has not been cloned "
                                    f"into the destination space, so clone it first in the destination space")
                        dict_list[key] = self.__clone_item_to_space(item_type=self.__src_id_type_dict.get(value),
                                                                    item_id=value)
                        logger.info(f"the reference value {value} was cloned and replaced with {dict_list.get(key)}")
                    # remove the broken reference ids
                    elif self.__check_broken_item_id(string=value):
                        logger.warning(f"the reference value {value} is a broken reference id, so assign null to it")
                        dict_list[key] = None
                else:
                    self.__replace_ids(dict_list=value)
                if isinstance(key, str):
                    if key in self.__src_id_vs_dst_id_dict:
                        new_key = self.__src_id_vs_dst_id_dict.get(key)
                        dict_list[new_key] = dict_list.pop(key)
                        logger.info(f"the reference key {key} has already been created in the destination space "
                                    f"and was replaced with {new_key}")
                    elif key in self.__src_id_payload_dict:
                        logger.info(f"the reference key {key} exists in the source space but has not been cloned "
                                    f"into the destination space, so clone it first in the destination space")
                        new_key = self.__clone_item_to_space(item_type=self.__src_id_type_dict.get(key), item_id=key)
                        dict_list[new_key] = dict_list.pop(key)
                        logger.info(f"the reference key {key} was cloned and replaced with {new_key}")
                    # remove the broken reference ids
                    elif self.__check_broken_item_id(string=key):
                        logger.warning(f"Key {value} is a broken reference id, so pop it")
                        dict_list.pop(key)
                else:
                    self.__replace_ids(dict_list=key)
        elif isinstance(dict_list, list):
            logger.info(f"dict_list is a list, replace ids or delete broken ids")
            # we must do reversely to avoid unexpected result on deleting by index
            for index in range(len(dict_list) - 1, -1, -1):
                element = dict_list[index]
                if isinstance(element, str):
                    if element in self.__src_id_vs_dst_id_dict:
                        dict_list[index] = self.__src_id_vs_dst_id_dict.get(element)
                        logger.info(f"the reference element {element} has already been created in the destination "
                                    f"space and was replaced with {dict_list[index]}")
                    elif element in self.__src_id_payload_dict:
                        logger.info(f"the reference element {element} exists in the source space but has not been "
                                    f"cloned into the destination space, so clone it first in the destination space")
                        dict_list[index] = self.__clone_item_to_space(item_type=self.__src_id_type_dict.get(element),
                                                                      item_id=element)
                        logger.info(f"the reference element {element} was cloned and replaced with {dict_list[index]}")
                    elif self.__check_broken_item_id(string=element):
                        logger.warning(f"element {element} is a broken reference id, so delete it; index {index}")
                        del dict_list[index]
                else:
                    self.__replace_ids(dict_list=element)
        # None, boolean, integer, float etc
        else:
            logger.info(f"the type is {type(dict_list)} and value is {dict_list}; skip it")

    def __clone_child(self, src_parent_id=None, dst_parent_id=None, child_type=None, child_id_key=None):
        # source item
        parent_type = self.__src_id_type_dict.get(src_parent_id)
        src_parent = self.__src_id_payload_dict.get(src_parent_id)
        src_child_id = src_parent.get(child_id_key)
        parent_name = src_parent.get(name_key)

        # destination item
        dst_parent = self.__dst_id_payload_dict.get(dst_parent_id)
        dst_child_id = dst_parent.get(child_id_key)
        self.__src_id_vs_dst_id_dict[src_child_id] = dst_child_id

        logger.info(f"clone {parent_type} {parent_name}'s {child_id_key} {src_child_id} from {self.__src_space_id} to "
                    f"{self.__dst_space_id}")
        src_child = self.__src_id_payload_dict.get(src_child_id)
        # TODO Octopus bug https://help.octopus.com/t/504-gateway-time-out-on-getting-all-variables/24732
        if not src_child:
            logger.warning(f"{src_child_id} does not exist in the memory, so get it from {self.__src_space_id}")
            src_child = get_single_item_by_name_or_id(item_type=child_type, item_id=src_child_id,
                                                      space_id=self.__src_space_id)
            self.__src_id_payload_dict[src_child_id] = src_child
        dst_child = get_or_delete_single_item_by_id(item_type=child_type, item_id=dst_child_id,
                                                    space_id=self.__dst_space_id)
        src_child_copy = copy.deepcopy(src_child)
        src_child_copy[version_key] = dst_child.get(version_key)
        self.__replace_ids(dict_list=src_child_copy)
        # an Octopus bug: "PUT" variables does not return ScopeValues;
        # so have to run "GET" again after "PUT" to save values
        put_single_item(item_type=child_type, payload=src_child_copy, space_id=self.__dst_space_id)
        dst_child = get_single_item_by_name_or_id(item_type=child_type, item_id=dst_child_id,
                                                  space_id=self.__dst_space_id)
        self.__dst_id_payload_dict[dst_child_id] = dst_child
        return dst_child_id

    # deployment processes are not really created as new; once a project is created, the deployment process is
    # created automatically; so we need to copy the deployment processes from the source space into the destination
    # space matching the same projects; the API call is "PUT" not "POST"
    # same for the variables
    def __post_process_project(self, src_id=None, dst_id=None):
        logger.info(f"clone {item_type_deployment_processes} from {item_type_projects} {src_id} to {dst_id} in "
                    f"{self.__dst_space_id}")
        self.__clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_deployment_processes,
                           child_id_key=deployment_process_id_key)
        logger.info(f"clone {item_type_variables} from {item_type_projects} {src_id} to {dst_id} in "
                    f"{self.__dst_space_id}")
        self.__clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_variables,
                           child_id_key=variable_set_id_key)

    def __post_process_library_variable_set(self, src_id=None, dst_id=None):
        logger.info(f"clone {item_type_variables} from {item_type_library_variable_sets} {src_id} to {dst_id} in "
                    f"{self.__dst_space_id}")
        self.__clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_variables,
                           child_id_key=variable_set_id_key)

    def __post_process_runbook(self, src_id=None, dst_id=None):
        logger.info(f"clone {item_type_runbook_processes} from {item_type_runbooks} {src_id} to {dst_id} in "
                    f"{self.__dst_space_id}")
        self.__clone_child(src_parent_id=src_id, dst_parent_id=dst_id, child_type=item_type_runbook_processes,
                           child_id_key=runbook_process_id_key)

    # tenant variables is special and its id is also tenant id, such as "Tenants-401"
    # so you have to use a separate map to store the tenants variables
    # also put/post tenant variables uses a different url from put variables
    def __post_process_tenant_variables(self, src_id=None, dst_id=None):
        logger.info(f"clone {item_type_tenant_variables} from {item_type_tenants} {src_id} in {self.__src_space_id} "
                    f"to {dst_id} in {self.__dst_space_id}")
        src_tenant_variables = self.__src_tenant_variables_payload_dict.get(src_id)
        src_tenant_variables_copy = copy.deepcopy(src_tenant_variables)

        self.__replace_ids(dict_list=src_tenant_variables_copy)

        dst_tenant_variables = put_post_tenant_variables_save(tenant_id=dst_id, space_id=self.__dst_space_id,
                                                              tenant_variables=src_tenant_variables_copy)

        self.__dst_tenant_variables_payload_dict[dst_id] = dst_tenant_variables

    def __full_process_tags(self, item_type=None, item_name=None, item_id=None):
        logger.info(f"full process - clone {item_type} {item_name} {item_id} to {self.__dst_space_id}")
        tag_set_name = item_id.split(slash_sign)[0]
        src_list_tag_sets = self.__type_src_list_items_dict.get(item_type_tag_sets)
        src_tag_set = find_item(lst=src_list_tag_sets, key=name_key, value=tag_set_name)
        src_tag = self.__src_id_payload_dict.get(item_id)
        src_tag_copy = copy.deepcopy(src_tag)
        src_tag_copy.pop(id_key, None)
        dst_tag_set = self.__find_matched_dst_item_by_src_item(src_item_with_dst_ids=src_tag_set,
                                                               item_type=item_type_tag_sets)
        if dst_tag_set:
            logger.info(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} exists in {self.__dst_space_id}, "
                        f"so try to find {item_id}")
            if not find_item(lst=dst_tag_set.get(tags_key), key=canonical_tag_name_key, value=item_id):
                logger.info(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} does not have {item_id} in "
                            f"{self.__dst_space_id}, add it")
                dst_tag_set.get(tags_key).append(src_tag_copy)
                dst_tag_set = put_single_item_save(item_type=item_type_tag_sets, payload=dst_tag_set,
                                                   space_id=self.__dst_space_id,
                                                   overwrite=True)
            else:
                logger.info(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} already has {item_id} in "
                            f"{self.__dst_space_id}, skip")
        else:
            logger.info(f"{item_type_tag_sets} {dst_tag_set.get(name_key)} does not exist in {self.__dst_space_id}, "
                        f"so create it with {item_type_tags} {item_id}")
            dst_tag_set = copy.deepcopy(src_tag_set)
            dst_tag_set.pop(id_key, None)
            dst_tag_set[tags_key] = [src_tag_copy]
            dst_tag_set = post_single_item_save(item_type=item_type_tag_sets, payload=dst_tag_set,
                                                space_id=self.__dst_space_id)
        self.__src_id_vs_dst_id_dict[item_id] = item_id
        self.__dst_id_payload_dict[item_id] = find_item(lst=dst_tag_set.get(tags_key), key=canonical_tag_name_key,
                                                        value=item_id)
        return item_id

    def __load_types(self, fake_space=False):
        if fake_space:
            log_info_print(local_logger=logger,
                           msg=f"Reading files {self.__item_types_set} from fake space {self.__src_space_id}...")
        else:
            log_info_print(local_logger=logger,
                           msg=f"Downloading {self.__item_types_set} from space {self.__src_space_id}...")
        actual_src_space_id = None
        for item_type in self.__item_types_set:
            log_info_print(local_logger=logger, msg=f"loading type {item_type} from space {self.__src_space_id}")
            # for cloning space from another Octopus server
            if fake_space:
                logger.info(f"Loading {item_type} in source fake space {self.__src_space_id} from the local file...")
                src_list_items = get_list_items_from_file(item_type=item_type, space_id=self.__src_space_id)
            # for cloning space to space on the same Octopus server
            else:
                logger.info(f"Loading {item_type} from source space {self.__src_space_id}...")
                src_list_items = get_one_type_to_list(item_type=item_type, space_id=self.__src_space_id)
            self.__type_src_list_items_dict[item_type] = src_list_items
            for src_item in src_list_items:
                if not actual_src_space_id and fake_space and item_type == item_type_projects \
                        and src_item.get(space_id_key):
                    actual_src_space_id = src_item.get(space_id_key)
                    logger.info(f"the actual source space id is {actual_src_space_id}; the fake one is "
                                f"{self.__src_space_id}")
                if src_item.get(id_key):
                    self.__src_id_payload_dict[src_item.get(id_key)] = src_item
                    self.__src_id_type_dict[src_item.get(id_key)] = item_type
                elif src_item.get(tenant_id_key):
                    self.__src_tenant_variables_payload_dict[src_item.get(tenant_id_key)] = src_item
                else:
                    log_raise_value_error(local_logger=logger, item=src_item, err=f"{item_type} does not have valid id")
        if fake_space and not actual_src_space_id:
            log_raise_value_error(local_logger=logger, err=f"Could not find an actual space id inside the fake space "
                                                           f"{self.__src_space_id}")
        return actual_src_space_id if actual_src_space_id else self.__src_space_id

    def __prep_tag_sets(self):
        log_info_print(local_logger=logger, msg="prepare for all tags")
        src_list_tag_sets = self.__type_src_list_items_dict.get(item_type_tag_sets)
        for src_tag_set in src_list_tag_sets:
            for src_tag in src_tag_set.get(tags_key):
                canonical_tag_name = src_tag.get(canonical_tag_name_key)
                self.__src_id_payload_dict[canonical_tag_name] = src_tag
                self.__src_id_type_dict[canonical_tag_name] = item_type_tags

    def __initialize_maps(self, src_space_id=None, dst_space_id=None, fake_space=False):
        log_info_print(local_logger=logger,
                       msg=f"initialize the map for cloning from space {src_space_id} to {dst_space_id}; the source "
                           f"space is {'fake' if fake_space else 'real'}")
        self.__src_space_id = src_space_id
        self.__dst_space_id = dst_space_id

        self.__type_prep_func_dict[item_type_tag_sets] = self.__prepare_tag_set
        self.__type_prep_func_dict[item_type_projects] = self.__prepare_project
        self.__type_prep_func_dict[item_type_library_variable_sets] = self.__prepare_library_variable_set
        self.__type_prep_func_dict[item_type_feeds] = self.__prepare_feed
        self.__type_prep_func_dict[item_type_runbooks] = self.__prepare_runbook
        self.__type_prep_func_dict[item_type_accounts] = self.__prepare_account

        self.__type_post_func_dict[item_type_projects] = self.__post_process_project
        self.__type_post_func_dict[item_type_library_variable_sets] = self.__post_process_library_variable_set
        self.__type_post_func_dict[item_type_tenants] = self.__post_process_tenant_variables
        self.__type_post_func_dict[item_type_runbooks] = self.__post_process_runbook

        self.__type_full_func_dict[item_type_tags] = self.__full_process_tags

        actual_src_space_id = self.__load_types(fake_space=fake_space)

        if fake_space:
            self.__src_id_vs_dst_id_dict[actual_src_space_id] = dst_space_id
        else:
            self.__src_id_vs_dst_id_dict[src_space_id] = dst_space_id

        self.__prep_tag_sets()

    def __save_space_map(self):
        log_info_print(local_logger=logger, msg=f"")
        current_time = strftime("%Y-%m-%d-%H-%M-%S", gmtime())
        local_file = get_local_single_item_file(item_name=space_map + underscore_sign + current_time,
                                                item_type=item_type_migration, space_id=self.__dst_space_id)
        log_info_print(local_logger=logger, msg=f"writing item id map to {local_file}...")
        save_file(file_path_name=local_file, content=self.__src_id_vs_dst_id_dict)

    def clone_space(self, src_space_id=None, dst_space_id=None, item_types_comma_delimited=None, fake_space=False):
        if item_types_comma_delimited:
            process_types = must_have_types + item_types_comma_delimited.split(comma_sign)
        else:
            process_types = normal_cloneable_types
        log_info_print(local_logger=logger, msg=f"cloning {process_types} from {src_space_id} to {dst_space_id}...")
        if not config.overwrite:
            config.overwrite = input(f"***** You are cloning {process_types} from {src_space_id} to {dst_space_id}; "
                                     f"Some entities may already exist in {dst_space_id}; "
                                     f"Do you want to overwrite the existing entities? "
                                     f"If no, we will skip the existing entities. [Y/n]: ") == 'Y'
        self.__item_types_set = set(normal_cloneable_types + process_types +
                                    [item_type_deployment_processes, item_type_variables, item_type_tenant_variables])
        self.__initialize_maps(src_space_id=src_space_id, dst_space_id=dst_space_id, fake_space=fake_space)
        for item_type in process_types:
            if item_type in normal_cloneable_types:
                self.__clone_type_to_space(item_type=item_type)
        self.__save_space_map()

    def clone_space_item(self, src_space_id=None, dst_space_id=None, item_type=None, item_name=None, item_id=None,
                         fake_space=False):
        item_badge = item_name if item_name else item_id
        log_info_print(local_logger=logger,
                       msg=f"cloning {item_type} {item_badge} from {src_space_id} to {dst_space_id}...")
        if not config.overwrite:
            config.overwrite = input(
                f"***** You are cloning {item_type} {item_badge} from {src_space_id} to {dst_space_id}; "
                f"Some entities may already exist in {dst_space_id}; "
                f"Do you want to overwrite the existing entities? "
                f"If no, we will skip the existing entities. [Y/n]: ") == 'Y'
        self.__item_types_set = set(normal_cloneable_types + [item_type, item_type_deployment_processes,
                                                              item_type_variables, item_type_tenant_variables])
        self.__initialize_maps(src_space_id=src_space_id, dst_space_id=dst_space_id, fake_space=fake_space)
        for must_have_type in must_have_types:
            self.__clone_type_to_space(item_type=must_have_type)
        if item_type in normal_cloneable_types:
            self.__clone_item_to_space(item_type=item_type, item_name=item_name, item_id=item_id)
        self.__save_space_map()


if __name__ == "__main__":
    migration = Migration()



import copy
import logging

from octopus_python_client.common import get_one_type_save, item_type_deployment_processes, get_child_item_save, \
    item_type_projects, update_child_item_from_local_save, clone_child_item_from_another_parent_save, get_child_item, \
    put_child_item_save, deployment_process_id_key, steps_key, name_key, id_key, actions_key, log_info_print
from octopus_python_client.utilities.helper import find_index, find_item, log_raise_value_error

logger = logging.getLogger(__name__)


def get_all_deployment_processes(space_id=None):
    return get_one_type_save(item_type=item_type_deployment_processes, space_id=space_id)


def get_deployment_process(project_literal_name=None, space_id=None):
    return get_child_item_save(parent_name=project_literal_name, parent_type=item_type_projects,
                               child_id_key=deployment_process_id_key, child_type=item_type_deployment_processes,
                               space_id=space_id)


def update_deployment_process(project_literal_name=None, space_id=None):
    update_child_item_from_local_save(parent_name=project_literal_name, parent_type=item_type_projects,
                                      child_id_key=deployment_process_id_key, child_type=item_type_deployment_processes,
                                      space_id=space_id)


# clone deployment process from a base project to a project
def clone_deployment_process(project_literal_name=None, base_project_name=None, space_id=None):
    clone_child_item_from_another_parent_save(parent_name=project_literal_name, base_parent_name=base_project_name,
                                              parent_type=item_type_projects, child_id_key=deployment_process_id_key,
                                              child_type=item_type_deployment_processes, sub_item_key=steps_key,
                                              space_id=space_id)


def clone_process_step(project_literal_name=None, step_name=None, base_step_name=None, prev_step_name=None,
                       space_id=None):
    if not project_literal_name or not step_name or not base_step_name:
        raise ValueError('Project literal name or step or base step names must not be empty')
    log_info_print(local_logger=logger,
                   msg=f"clone project {project_literal_name} step from base step {base_step_name} to new step "
                       f"{step_name} and place it after step {prev_step_name}")
    process = get_child_item(parent_name=project_literal_name, parent_type=item_type_projects,
                             child_id_key=deployment_process_id_key, child_type=item_type_deployment_processes,
                             space_id=space_id)
    steps = process[steps_key]
    step = copy.deepcopy(find_item(lst=steps, key=name_key, value=base_step_name))
    step[name_key] = step_name
    step[id_key] = ""
    for action in step[actions_key]:
        action[name_key] = step_name
        action[id_key] = ""
    if prev_step_name:
        prev_step_name_index = find_index(lst=steps, key=name_key, value=prev_step_name)
        if prev_step_name_index < 0:
            log_raise_value_error(local_logger=logger, err=f"previous step {prev_step_name} does not exist")
        steps.insert(prev_step_name_index + 1, step)
    else:
        steps.append(step)
    child_item = put_child_item_save(parent_name=project_literal_name, child_type=item_type_deployment_processes,
                                     payload=process, space_id=space_id)
    return child_item


def delete_process_step(project_literal_name=None, step_name=None, space_id=None):
    if not project_literal_name or not step_name:
        raise ValueError('Project literal name or step name must not be empty')
    log_info_print(local_logger=logger,
                   msg=f"deleting step {step_name} of project {project_literal_name} in space {space_id}")
    process = get_child_item(parent_name=project_literal_name, parent_type=item_type_projects,
                             child_id_key=deployment_process_id_key, child_type=item_type_deployment_processes,
                             space_id=space_id)
    steps = process[steps_key]
    step_index = find_index(lst=steps, key=name_key, value=step_name)
    if step_index < 0:
        log_raise_value_error(local_logger=logger, err=f"step {step_name} does not exit")
    steps.pop(step_index)
    child_item = put_child_item_save(parent_name=project_literal_name, child_type=item_type_deployment_processes,
                                     payload=process, space_id=space_id)
    return child_item



import logging
import sys
from types import SimpleNamespace

import requests

from octopus_python_client.utilities.helper import log_raise_value_error

operation_delete = "delete"
operation_get = "get"
operation_post = "post"
operation_put = "put"

application_json = "application/json"
content_type_key = "Content-Type"
login_payload_password_key = "Password"
login_payload_user_name_key = "Username"
users_login_url_suffix = "users/login"
x_octopus_api_key_key = "X-Octopus-ApiKey"

logger = logging.getLogger(__name__)


def call_octopus(config=None, url_suffix=None, operation=None, payload=None):
    url = config.octopus_endpoint + url_suffix if url_suffix else config.octopus_endpoint
    operation = operation if operation else operation_get
    with requests.Session() as session:
        if config.api_key:
            headers = {content_type_key: application_json, x_octopus_api_key_key: config.api_key}
        elif config.user_name and config.password:
            headers = {content_type_key: application_json}
            login_url = config.octopus_endpoint + users_login_url_suffix
            login_payload = {login_payload_user_name_key: config.user_name, login_payload_password_key: config.password}
            session.post(login_url, json=login_payload, headers=headers)
        else:
            log_raise_value_error(local_logger=logger, err=f"either api_key or user_name and password are required")
        try:
            logger.info(f"{operation}: " + url)
            if operation.lower() == operation_post:
                session_response = session.post(url, json=payload, headers=headers)
            elif operation.lower() == operation_get:
                session_response = session.get(url, params=payload, headers=headers)
            elif operation.lower() == operation_put:
                session_response = session.put(url, json=payload, headers=headers)
            elif operation.lower() == operation_delete:
                session_response = session.delete(url, headers=headers)
            else:
                log_raise_value_error(local_logger=logger, err=f"Invalid operation: {operation}; only post, get, put "
                                                               f"and delete are supported")
            logger.info("response status code: " + str(session_response.status_code))
            response_json = ""
            if session_response.text:
                response_json = session_response.json()
            # if permission is denied, continue with other operations
            if session_response.status_code == 403:
                logger.warning(response_json)
            elif session_response.status_code < 200 or session_response.status_code > 299:
                log_raise_value_error(local_logger=logger, err=response_json)
            return response_json
        except requests.exceptions.RequestException as e:
            logger.error(e)
            sys.exit(1)


# TODO for testing purpose, to be removed
if __name__ == "__main__":
    octopus_config = {"octopus_endpoint": "https://demo.octopusdeploy.com/api/", "api_key": None, "user_name": "guest",
                      "password": "guest"}
    response = call_octopus(config=SimpleNamespace(**octopus_config), url_suffix="Spaces-1/environments")
    logger.info(response)



import logging
from pprint import pformat

from octopus_python_client.common import request_octopus_item, item_type_deployment_processes, verify_space, \
    item_type_projects, get_single_item_by_name, id_key, deployment_process_id_key, item_type_channels, \
    find_sub_by_item, packages_key, action_name_key, package_reference_name_key, feed_id_key, package_id_key, \
    item_type_feeds, item_type_packages, version_key, items_key, get_list_variables_by_set_name_or_id, name_key, \
    value_key, project_id_key, next_version_increment_key, release_notes_key, channel_id_key, selected_packages_key, \
    item_type_releases, save_single_item, item_type_deployments, item_type_tenants, item_type_environments, \
    get_item_id_by_name, tenant_id_key, environment_id_key, release_id_key, comments_key, log_info_print
from octopus_python_client.utilities.helper import replace_list_new_value, get_dict_from_str
from octopus_python_client.utilities.send_requests_to_octopus import operation_post

logger = logging.getLogger(__name__)


class ReleaseDeployment:
    def __init__(self, project_name=None, channel_name=None, notes=None, space_id_name=None, package_version_dict=None,
                 packages_variable_set_name=None):
        assert space_id_name, "space name/id must not be empty!"
        assert project_name, "project name must not be empty!"

        self._space_id = verify_space(space_id_name=space_id_name)
        assert self._space_id, f"Space {space_id_name} cannot be found or you do not have permission to access!"

        project = get_single_item_by_name(item_type=item_type_projects, item_name=project_name,
                                          space_id=self._space_id)
        assert project, f"Project {project_name} cannot be found or you do not have permission to access!"
        self._project_id = project.get(id_key)
        self._deployment_process_id = project.get(deployment_process_id_key)
        self._release_request_payload = {project_id_key: self._project_id, release_notes_key: notes}

        self._channel_id = ""
        if channel_name:
            channel = find_sub_by_item(item_type=item_type_projects, item_id=self._project_id,
                                       sub_type=item_type_channels, sub_name=channel_name, space_id=self._space_id)
            assert channel, f"Cannot find channel {channel_name} in project {project_name}"
            self._channel_id = channel.get(id_key)
        if self._channel_id:
            self._release_request_payload[channel_id_key] = self._channel_id

        # package versions read from a variable set, e.g. {"Name": "package.near", "Value": "20.0225.1714"}
        self._packages_variable_set_name = packages_variable_set_name

        # user selected package versions, e.g. "{'package.near': '20.0225.1714'}"
        self._package_version_dict = None
        if package_version_dict:
            logger.info(f"converting package version {package_version_dict} to dict...")
            self._package_version_dict = get_dict_from_str(string=package_version_dict)

        self._template = None
        self._selected_packages = None
        self._release_response = None
        self._release_id = None

    def _get_deployment_process_template(self):
        logger.info(f"Fetching deployment process template from {self._deployment_process_id} with channel "
                    f"{self._channel_id} in {self._space_id}, which is used for defining release and deployment")
        address = f"{item_type_deployment_processes}/{self._deployment_process_id}/template?channel=" \
                  f"{self._channel_id}"
        self._template = request_octopus_item(space_id=self._space_id, address=address)

    def _get_selected_packages(self):
        logger.info(f"getting package information for each step")
        self._selected_packages = []
        for package in self._template.get(packages_key):
            logger.info(f"getting package information for {package.get(action_name_key)} with package "
                        f"{package.get(package_reference_name_key)}")
            address = f"{item_type_feeds}/{package.get(feed_id_key)}/{item_type_packages}/versions?packageId=" \
                      f"{package.get(package_id_key)}&take=1"
            package_detail = request_octopus_item(space_id=self._space_id, address=address)
            selected_package = {action_name_key: package.get(action_name_key),
                                package_reference_name_key: package.get(package_reference_name_key),
                                version_key: package_detail.get(items_key)[0].get(version_key)}
            self._selected_packages.append(selected_package)

    def _update_package_version(self, package=None, version=None):
        match_dict = {package_reference_name_key: package}
        replace_dict = {version_key: version}
        replace_list_new_value(lst=self._selected_packages, match_dict=match_dict, replace_dict=replace_dict)

    def _update_selected_packages(self):
        if self._packages_variable_set_name:
            list_release_versions = get_list_variables_by_set_name_or_id(set_name=self._packages_variable_set_name,
                                                                         space_id=self._space_id)
            for package_version in list_release_versions:
                self._update_package_version(package=package_version.get(name_key),
                                             version=package_version.get(value_key))
        if self._package_version_dict:
            for package, version in self._package_version_dict.items():
                self._update_package_version(package=package, version=version)

    # release version must be unique for each release
    def create_release(self, release_version=None):
        self._get_deployment_process_template()
        self._get_selected_packages()
        self._update_selected_packages()
        if not release_version:
            release_version = self._template.get(next_version_increment_key)
        self._release_request_payload[version_key] = release_version
        self._release_request_payload[selected_packages_key] = self._selected_packages
        logger.info("the request release payload is")
        logger.info(pformat(self._release_request_payload))
        self._release_response = request_octopus_item(payload=self._release_request_payload, space_id=self._space_id,
                                                      address=item_type_releases, action=operation_post)
        logger.info("the response release payload is")
        logger.info(pformat(self._release_response))
        save_single_item(item_type=item_type_releases, item=self._release_response, space_id=self._space_id)
        self._release_id = self._release_response.get(id_key)
        return self._release_response

    @staticmethod
    def create_deployment_direct(release_id=None, environment_name=None, tenant_name=None, space_id_name=None,
                                 comments=None):
        logger.info(f"creating a deployment for {release_id} in space {space_id_name} with environment "
                    f"{environment_name}, tenant {tenant_name} and comments: {comments}")

        assert release_id, "release_id must not be empty!"
        assert space_id_name, "space id/name must not be empty!"
        assert environment_name, "environment_name must not be empty!"
        assert tenant_name, "tenant_name must not be empty!"

        space_id = verify_space(space_id_name=space_id_name)
        assert space_id, f"Space {space_id_name} cannot be found or you do not have permission to access!"

        deployment_request_payload = {release_id_key: release_id,
                                      environment_id_key: get_item_id_by_name(item_type=item_type_environments,
                                                                              item_name=environment_name,
                                                                              space_id=space_id),
                                      tenant_id_key: get_item_id_by_name(item_type=item_type_tenants,
                                                                         item_name=tenant_name,
                                                                         space_id=space_id),
                                      comments_key: comments}
        logger.info("the request deployment payload is")
        logger.info(pformat(deployment_request_payload))
        deployment_response_payload = request_octopus_item(payload=deployment_request_payload, space_id=space_id,
                                                           address=item_type_deployments, action=operation_post)
        logger.info("the response deployment payload is")
        log_info_print(local_logger=logger, msg=deployment_response_payload)
        save_single_item(item_type=item_type_deployments, item=deployment_response_payload, space_id=space_id)
        return deployment_response_payload

    def create_deployment_for_current_release(self, environment_name=None, tenant_name=None, comments=None):
        return self.create_deployment_direct(
            release_id=self._release_id, environment_name=environment_name, tenant_name=tenant_name,
            space_id_name=self._space_id, comments=comments)

    @property
    def release_id(self):
        logger.info(f"get the release id {self._release_id}")
        return self._release_id

    @property
    def release_response(self):
        logger.info(f"get the release response for {self._release_id}")
        return self._release_response

    @staticmethod
    def create_release_direct(release_version=None, project_name=None, channel_name=None, notes=None,
                              space_id_name=None, package_version_json=None, packages_variable_set_name=None):
        release = ReleaseDeployment(
            project_name=project_name, channel_name=channel_name, notes=notes, space_id_name=space_id_name,
            package_version_dict=package_version_json, packages_variable_set_name=packages_variable_set_name)
        release.create_release(release_version=release_version)
        log_info_print(local_logger=logger, msg=release.release_response)
        return release

    @staticmethod
    def create_release_deployment(release_version=None, project_name=None, channel_name=None, notes=None,
                                  space_id_name=None, package_version_json=None, packages_variable_set_name=None,
                                  environment_name=None, tenant_name=None, comments=None):
        release = ReleaseDeployment.create_release_direct(
            release_version=release_version, project_name=project_name, channel_name=channel_name, notes=notes,
            space_id_name=space_id_name, package_version_json=package_version_json,
            packages_variable_set_name=packages_variable_set_name)
        return release.create_deployment_for_current_release(environment_name=environment_name, tenant_name=tenant_name,
                                                             comments=comments)



import argparse
import logging
import os

from octopus_python_client.common import config, get_one_type_save, get_single_item_by_name_or_id_save, \
    update_single_item_save, create_single_item_from_local_file, item_type_deployment_processes, \
    clone_single_item_from_remote_item, delete_single_item_by_name_or_id, get_child_item_save, \
    update_child_item_from_local_save, clone_child_item_from_another_parent_save, get_types_save, \
    item_types_only_ourter_space, item_types_inside_space, deployment_process_id_key, steps_key, \
    merge_single_item_save, delete_one_type, api_key_key, octopus_endpoint_key, octopus_name_key, user_name_key, \
    password_key, double_hyphen, verify_space, get_spaces_save, get_task_status, wait_task
from octopus_python_client.migration import Migration
from octopus_python_client.processes import clone_process_step, delete_process_step
from octopus_python_client.projects import clone_project, delete_project, get_project, project_update_variable_sets
from octopus_python_client.release_deployment import ReleaseDeployment
from octopus_python_client.utilities.helper import log_raise_value_error

logging.basicConfig(filename=os.path.join(os.getcwd(), "octopus_python_client.log"),
                    filemode='a',
                    format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)
logger = logging.getLogger(__name__)


class Actions:
    action_get_spaces = "get_spaces"  # get all types
    action_get_types = "get_types"  # get all types
    action_get_type = "get_type"  # get all items under one type
    action_delete_type = "delete_type"  # delete all items under one type
    action_get = "get"  # get one item
    action_update = "update"  # update one item
    action_update_merge = "update_merge"  # update one item and merge the existing sub-items
    action_create = "create"  # create a new item
    action_clone = "clone"  # clone a new item
    action_delete = "delete"  # delete an item
    action_get_child = "get_child"  # get a child item
    action_update_child = "update_child"  # update a child item
    action_clone_child = "clone_child"  # clone a child item
    action_clone_process_step = "clone_process_step"  # clone a project process step
    action_delete_process_step = "delete_process_step"  # delete a project process step
    action_clone_project = "clone_project"  # clone a project including process
    action_delete_project = "delete_project"  # delete a project including process
    action_get_project = "get_project"  # get a project including process
    action_project_update_variable_sets = "update_variable_sets"  # update the variable sets for a project
    # clone Octopus single item from one space to another space
    action_clone_space_item = "clone_space_item"
    # clone a few types from one space to another space
    action_clone_space = "clone_space"
    action_task_status = "task_status"
    action_wait_task = "wait_task"
    action_create_release = "create_release"
    action_create_deployment = "create_deployment"
    action_create_release_deployment = "create_release_deployment"


clone_space_actions = {Actions.action_clone_space_item, Actions.action_clone_space}


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("-o", double_hyphen + octopus_endpoint_key, help="octopus endpoint")
    parser.add_argument("-s", "--space_id_name", help="octopus space id or name")
    parser.add_argument("-ss", "--spaces",
                        help='list of octopus space id or name, like "my_space,Spaces-1,Spaces-2"')
    parser.add_argument("-ds", "--dst_space_id_name", help="destination octopus space id or name for clone/migration")
    parser.add_argument("-n", double_hyphen + octopus_name_key,
                        help="customized octopus server name, used for folder name")
    parser.add_argument("-k", double_hyphen + api_key_key,
                        help="api key for octopus; either api_key or user_name and password are required")
    parser.add_argument("-u", double_hyphen + user_name_key,
                        help="user_name for octopus; either api_key or user_name and password are required")
    parser.add_argument("-p", double_hyphen + password_key,
                        help="password for octopus; either api_key or user_name and password are required")
    parser.add_argument("-a", "--action", help=str(Actions.__dict__.values()), required=True)
    parser.add_argument("-ow", "--overwrite", help="if present, overwrite = True", action='store_true')
    parser.add_argument("-ns", "--no_stdout", help="if present, no_stdout = True, means no stdout", action='store_true')
    parser.add_argument("-ts", "--item_types",
                        help='if not item_types and not octopus_space_id, get all item types '
                             'regardless whether they are above Spaces; if (not item_types) and octopus_space_id, '
                             'get all item types below octopus_space_id; '
                             'list like "accounts,actiontemplates,artifacts" is also accepted; '
                             'item types above Spaces: ' + ", ".join(item_types_only_ourter_space) +
                             "; \nitem types above and under Spaces: " + ", ".join(item_types_inside_space))
    parser.add_argument("-tp", "--item_type", help="one of item types above Spaces: " + ", ".join(
        item_types_only_ourter_space) + "; \nitem types above and under Spaces: " + ", ".join(item_types_inside_space))
    parser.add_argument("-nm", "--item_name", help="item name: e.g. project_name")
    parser.add_argument("-id", "--item_id", help="item id: e.g. Lifecycles-1")
    parser.add_argument("-bn", "--base_item_name", help="base item name, either local or on Octopus server")
    parser.add_argument("-pn", "--parent_name", help="parent item name: e.g. project_name")
    parser.add_argument("-pt", "--parent_type", help="parent item type: e.g. projects")
    parser.add_argument("-ct", "--child_type", help=f"child item type: e.g. {item_type_deployment_processes}")
    parser.add_argument("-ck", "--child_id_key", help=f"child id key: e.g. {deployment_process_id_key}")
    parser.add_argument("-sk", "--sub_item_key", help=f"sub-item key: this sub-item is for copy/clone e.g. {steps_key}")
    parser.add_argument("-bp", "--base_parent_name", help="base parent item name: e.g. a base project_name")
    parser.add_argument("-sn", "--step_name", help="step name as in octopus process")
    parser.add_argument("-bs", "--base_step_name", help="base step name as in octopus process for cloning a step")
    parser.add_argument("-ps", "--prev_step_name",
                        help="previous step name in octopus process for the step insertion location")
    parser.add_argument("-sf", "--suffix", help="variable sets name suffix")
    parser.add_argument("-rs", "--remove_suffix",
                        help="if present, remove suffix from variable sets name, otherwise add suffix",
                        action='store_true')
    parser.add_argument("-ti", "--task_id", help="task id, like ServerTasks-1234")
    parser.add_argument("-tl", "--time_limit_second", help="time limit in second")
    parser.add_argument("-rv", "--release_version", help="release version for creating a new release")
    parser.add_argument("-pj", "--project_name", help="project name")
    parser.add_argument("-cn", "--channel_name", help="channel name")
    parser.add_argument("-nt", "--notes", help="notes")
    parser.add_argument("-pv", "--package_version_json",
                        help="user selected package versions, e.g. \""
                             r"{\"package.near\": \"20.0225.1714\", \"package.main\": \"20.0325.0839\"}" "\"")
    parser.add_argument("-vs", "--variable_set_name", help="variable set name")
    parser.add_argument("-ri", "--release_id", help="release id for deployment")
    parser.add_argument("-en", "--environment_name", help="environment name, like Integration")
    parser.add_argument("-tn", "--tenant_name", help="tenant name, like cd-near")
    parser.add_argument("-cm", "--comments", help="comments")

    args, unknown = parser.parse_known_args()
    return args


def run():
    args = _parse_args()

    if args.octopus_endpoint:
        config.octopus_endpoint = args.octopus_endpoint
    if not config.octopus_endpoint:
        raise ValueError("octopus_endpoint must not be empty")

    if args.octopus_name:
        config.octopus_name = args.octopus_name
    if not config.octopus_name:
        raise ValueError("octopus_name must not be empty")

    if args.api_key:
        config.api_key = args.api_key
    if args.user_name:
        config.user_name = args.user_name
    if args.password:
        config.password = args.password
    if not config.api_key and not (config.user_name and config.password):
        raise ValueError(f"either api_key or user_name and password are required")

    if args.overwrite:
        config.overwrite = args.overwrite
    if args.no_stdout:
        config.no_stdout = args.no_stdout

    logger.info(config.__dict__)

    # verify space id/name
    dst_space_id = None
    if args.dst_space_id_name:
        dst_space_id = verify_space(space_id_name=args.dst_space_id_name)
        if dst_space_id:
            logger.info(f"destination space_id is: {dst_space_id}")
        else:
            raise ValueError(f"the destination space id/name {args.dst_space_id_name} you specified does not exist or "
                             f"you may not have permission to access it")

    space_id = None
    fake_space = False
    if args.space_id_name:
        space_id = verify_space(space_id_name=args.space_id_name)
        if space_id:
            logger.info(f"octopus space_id is: {space_id}")
        elif args.action not in clone_space_actions:
            raise ValueError(f"the space id/name {args.space_id_name} you specified does not exist "
                             f"or you may not have permission to access it")
        elif input(f"Are you sure you want to {args.action} from nonexistent space {args.space_id_name} [Y/n]?") == 'Y':
            space_id = args.space_id_name
            fake_space = True
        else:
            return
    elif args.action != Actions.action_get_spaces \
            and input(f"Are you sure you want to run a command against null space [Y/n]? ") != 'Y':
        return

    if args.action == Actions.action_get_spaces:
        get_spaces_save(item_types_comma_delimited=args.item_types, space_id_or_name_comma_delimited=args.spaces)
    elif args.action == Actions.action_get_types:
        get_types_save(item_types_comma_delimited=args.item_types, space_id=space_id)
    elif args.action == Actions.action_get_type:
        get_one_type_save(item_type=args.item_type, space_id=space_id)
    elif args.action == Actions.action_delete_type:
        delete_one_type(item_type=args.item_type, space_id=space_id)
    elif args.action == Actions.action_get:
        get_single_item_by_name_or_id_save(item_type=args.item_type, item_name=args.item_name, item_id=args.item_id,
                                           space_id=space_id)
    elif args.action == Actions.action_update:
        update_single_item_save(item_type=args.item_type, item_name=args.item_name, item_id=args.item_id,
                                space_id=space_id)
    elif args.action == Actions.action_update_merge:
        merge_single_item_save(item_type=args.item_type, item_name=args.item_name, item_id=args.item_id,
                               child_id_key=args.child_id_key, space_id=space_id)
    elif args.action == Actions.action_create:
        create_single_item_from_local_file(item_type=args.item_type, item_name=args.item_name,
                                           local_item_name=args.base_item_name, space_id=space_id)
    elif args.action == Actions.action_clone:
        clone_single_item_from_remote_item(item_type=args.item_type, item_name=args.item_name,
                                           base_item_name=args.base_item_name, space_id=space_id)
    elif args.action == Actions.action_delete:
        delete_single_item_by_name_or_id(item_type=args.item_type, item_name=args.item_name, item_id=args.item_id,
                                         space_id=space_id)
    elif args.action == Actions.action_get_child:
        get_child_item_save(parent_name=args.parent_name, parent_type=args.parent_type, child_id_key=args.child_id_key,
                            child_type=args.child_type, space_id=space_id)
    elif args.action == Actions.action_update_child:
        update_child_item_from_local_save(parent_name=args.parent_name, parent_type=args.parent_type,
                                          child_id_key=args.child_id_key, child_type=args.child_type, space_id=space_id)
    elif args.action == Actions.action_clone_child:
        clone_child_item_from_another_parent_save(
            parent_name=args.parent_name, base_parent_name=args.base_parent_name, parent_type=args.parent_type,
            child_id_key=args.child_id_key, child_type=args.child_type, sub_item_key=args.sub_item_key,
            space_id=space_id)
    elif args.action == Actions.action_clone_process_step:
        clone_process_step(project_literal_name=args.project_name, step_name=args.step_name,
                           base_step_name=args.base_step_name, prev_step_name=args.prev_step_name,
                           space_id=space_id)
    elif args.action == Actions.action_delete_process_step:
        delete_process_step(project_literal_name=args.project_name, step_name=args.step_name, space_id=space_id)
    elif args.action == Actions.action_clone_project:
        clone_project(project_literal_name=args.project_name, base_project_name=args.base_item_name,
                      space_id=space_id)
    elif args.action == Actions.action_delete_project:
        delete_project(project_literal_name=args.project_name, space_id=space_id)
    elif args.action == Actions.action_get_project:
        get_project(project_literal_name=args.project_name, space_id=space_id)
    elif args.action == Actions.action_project_update_variable_sets:
        project_update_variable_sets(project_literal_name=args.project_name, suffix=args.suffix, space_id=space_id,
                                     remove_suffix=args.remove_suffix)
    elif args.action == Actions.action_clone_space:
        Migration().clone_space(src_space_id=space_id, dst_space_id=dst_space_id,
                                item_types_comma_delimited=args.item_types, fake_space=fake_space)
    elif args.action == Actions.action_clone_space_item:
        Migration().clone_space_item(src_space_id=space_id, dst_space_id=dst_space_id, item_type=args.item_type,
                                     item_name=args.item_name, item_id=args.item_id, fake_space=fake_space)
    elif args.action == Actions.action_task_status:
        get_task_status(task_id=args.task_id, space_id=space_id)
    elif args.action == Actions.action_wait_task:
        wait_task(task_id=args.task_id, space_id=space_id, time_limit_second=args.time_limit_second)
    elif args.action == Actions.action_create_release:
        ReleaseDeployment.create_release_direct(
            release_version=args.release_version, project_name=args.project_name, channel_name=args.channel_name,
            notes=args.notes, space_id_name=space_id, package_version_json=args.package_version_json,
            packages_variable_set_name=args.variable_set_name)
    elif args.action == Actions.action_create_deployment:
        ReleaseDeployment.create_deployment_direct(
            release_id=args.release_id, environment_name=args.environment_name, tenant_name=args.tenant_name,
            space_id_name=space_id, comments=args.comments)
    elif args.action == Actions.action_create_release_deployment:
        ReleaseDeployment.create_release_deployment(
            release_version=args.release_version, project_name=args.project_name, channel_name=args.channel_name,
            notes=args.notes, space_id_name=space_id, package_version_json=args.package_version_json,
            packages_variable_set_name=args.variable_set_name, environment_name=args.environment_name,
            tenant_name=args.tenant_name, comments=args.comments)

    else:
        log_raise_value_error(local_logger=logger, err="We only support actions: " + str(Actions.__dict__.values()))


def main():
    logger.info("********** Octopus deploy python client tool - start **********")
    run()
    logger.info("********** Octopus deploy python client tool - end **********")


if __name__ == "__main__":
    main()



import copy
import logging
import os
import time
from pprint import pformat

from octopus_python_client.utilities.helper import compare_overwrite, find_item, compare_dicts, load_file, save_file
from octopus_python_client.utilities.send_requests_to_octopus import call_octopus, operation_get, operation_post, \
    operation_put, operation_delete

logger = logging.getLogger(__name__)

# constants
all_underscore = "all_"
error_message_key = "ErrorMessage"
error_message_resource_not_found = "The resource you requested was not found."
comma_sign = ","
dot_sign = "."
double_hyphen = "--"
environments_prefix = "Environments"
executing_string = "Executing"
file_configuration = "configuration.json"
folder_outer_spaces = "outer_spaces"
folder_configurations = "configurations"
hyphen_sign = "-"
runbook_process_prefix = "RunbookProcess"
slash_all = "/all"
slash_sign = "/"
space_map = "space_map"
success_string = "Success"
tenants_prefix = "Tenants"
underscore_sign = "_"
url_all_pages = "?skip=0&take=2147483647"
yaml_ext = ".yaml"
positive_integer_regex = "-[1-9][0-9]*$"

# dict keys
action_name_key = "ActionName"
actions_key = 'Actions'
api_key_key = "api_key"
canonical_tag_name_key = "CanonicalTagName"
channel_id_key = "ChannelId"
cloned_from_tenant_id_key = "ClonedFromTenantId"
comments_key = "Comments"
deployment_process_id_key = 'DeploymentProcessId'
description_key = "Description"
donor_package_key = "DonorPackage"
donor_package_step_id_key = "DonorPackageStepId"
environment_id_key = "EnvironmentId"
environment_ids_key = "EnvironmentIds"
feed_id_key = "FeedId"
file_name_key = "Filename"
id_key = 'Id'
included_library_variable_set_ids_key = "IncludedLibraryVariableSetIds"
items_key = 'Items'
life_cycle_id_key = "LifecycleId"
name_key = 'Name'
new_value_key = "NewValue"
next_version_increment_key = "NextVersionIncrement"
no_stdout_key = "no_stdout"
octopus_endpoint_key = "octopus_endpoint"
octopus_name_key = "octopus_name"
owner_id_key = "OwnerId"
overwrite_key = "overwrite"
package_id_key = "PackageId"
package_reference_name_key = "PackageReferenceName"
packages_key = "Packages"
password_key = "password"
project_group_id_key = "ProjectGroupId"
project_id_key = "ProjectId"
published_runbook_snapshot_id_key = "PublishedRunbookSnapshotId"
release_id_key = "ReleaseId"
release_notes_key = "ReleaseNotes"
runbook_id_key = "RunbookId"
runbook_process_id_key = "RunbookProcessId"
secret_key_key = "SecretKey"
selected_packages_key = "SelectedPackages"
state_key = "State"
steps_key = 'Steps'
space_id_key = "SpaceId"
tags_key = "Tags"
task_id_key = "TaskId"
team_id_key = "TeamId"
tenant_id_key = "TenantId"
token_key = "Token"
user_name_key = "user_name"
user_role_id_key = "UserRoleId"
value_key = "Value"
variables_key = "Variables"
variable_set_id_key = "VariableSetId"
version_key = "Version"
versioning_strategy_key = "VersioningStrategy"
worker_pool_id_key = "WorkerPoolId"

# item types
item_type_accounts = "accounts"
item_type_action_templates = "actiontemplates"
item_type_api_keys = "apikeys"
item_type_artifacts = "artifacts"
item_type_build_information = "build-information"
item_type_channels = "channels"
item_type_certificates = "certificates"
item_type_configuration = "configuration"
item_type_dashboard = "dashboard"
item_type_dashboard_configuration = "dashboardconfiguration"
item_type_dashboard_dynamic = "dashboard/dynamic"
item_type_deployments = "deployments"
item_type_deployment_processes = 'deploymentprocesses'
item_type_environments = "environments"
item_type_events = "events"
item_type_feeds = "feeds"
item_type_home = "home"
item_type_interruptions = "interruptions"
item_type_library_variable_sets = "libraryvariablesets"
item_type_life_cycles = "lifecycles"
item_type_machine_policies = "machinepolicies"
item_type_machine_roles = "machineroles"
item_type_machines = "machines"
item_type_packages = "packages"
item_type_permissions = "permissions"
item_type_permissions_configuration = "permissions/configuration"
item_type_permissions_export = "permissions/export"
item_type_project_groups = "projectgroups"
item_type_project_triggers = "projecttriggers"
item_type_projects = "projects"
item_type_proxies = "proxies"
item_type_releases = "releases"
item_type_runbook_processes = "runbookProcesses"
item_type_runbook_snapshots = "runbookSnapshots"  # clone requires package version, so not working now
item_type_runbooks = "runbooks"
item_type_scoped_user_roles = "scopeduserroles"  # the space id should not be null in the response, buggy!
item_type_spaces = 'spaces'
item_type_migration = "migration"
item_type_subscriptions = "subscriptions"
item_type_tag_sets = "tagsets"
item_type_tags = "tags"
item_type_tasks = "tasks"
item_type_teams = "teams"
item_type_tenants = "tenants"
item_type_tenant_variables = "tenantvariables"
item_type_user_onboarding = "useronboarding"
item_type_users = "users"
item_type_variables = "variables"
item_type_variables_names = "variables/names"
item_type_worker_pools = "workerpools"
item_type_workers = "workers"

# messages
remote_local_same_msg = "Remote and local items are the same and no change is needed, exit"

# some types have the extended types like: /api/users/{id}/permissions
user_ext_types = [item_type_api_keys, item_type_permissions, item_type_permissions_configuration,
                  item_type_teams]  # item_type_permissions_export is csv file
ext_types_map = {item_type_users: user_ext_types}

item_types_with_duplicate_names = \
    {item_type_channels, item_type_tasks, item_type_deployments, item_type_configuration, item_type_spaces}

item_types_without_single_item = \
    {item_type_dashboard, item_type_dashboard_dynamic, item_type_variables, item_type_variables_names}

# must have for a space to work
must_have_types = [item_type_environments]
# these types do not have dependencies on other types
basic_types = [item_type_action_templates, item_type_certificates, item_type_feeds, item_type_machine_policies,
               item_type_machines, item_type_proxies, item_type_subscriptions, item_type_tag_sets,
               item_type_teams, item_type_library_variable_sets, item_type_worker_pools]
# these types have links/dependencies on other types
complex_types = [item_type_workers, item_type_life_cycles, item_type_project_groups, item_type_projects,
                 item_type_runbooks, item_type_tenants, item_type_channels, item_type_project_triggers,
                 item_type_accounts, item_type_build_information]
# these types have only one item in a space; they are mostly space properties
space_single_item_types = [item_type_dashboard_configuration]
# these types are the child type of another type
child_types = [item_type_deployment_processes, item_type_runbook_processes]
# these types needs "/all" to get all items for this type
only_all_types_inside_space = [item_type_variables, item_type_tenant_variables, item_type_machine_roles]
# the other types not cloneable for now
other_types = [item_type_packages, item_type_releases, item_type_interruptions, item_type_user_onboarding,
               item_type_dashboard, item_type_dashboard_dynamic, item_type_deployments, item_type_variables_names,
               item_type_artifacts, item_type_home, item_type_scoped_user_roles, item_type_runbook_snapshots]
# too many items in them, so ignore for now
large_types = [item_type_tasks, item_type_events]
# the types which are cloneable
normal_cloneable_types = must_have_types + basic_types + complex_types
# the types live inside space
item_types_inside_space = normal_cloneable_types + child_types + only_all_types_inside_space + other_types
# the types live outside space (Octopus server types)
item_types_only_ourter_space = \
    ["authentication", "configuration/certificates", "communityactiontemplates", "externalsecuritygroupproviders",
     "featuresconfiguration", "letsencryptconfiguration", "licenses/licenses-current",
     "licenses/licenses-current-status", "maintenanceconfiguration", "octopusservernodes", "performanceconfiguration",
     "permissions/all", "scheduler", "serverconfiguration", "serverconfiguration/settings", "serverstatus",
     "smtpconfiguration", "smtpconfiguration/isconfigured", "upgradeconfiguration", item_type_users, "userroles",
     item_type_configuration, item_type_spaces]
# the sub item map for a specific type; this is for deleting the unused sub items when the item cannot be deleted
# e.g. tagsets: Tags is the key to get the list of the sub items; CanonicalTagName is for printing purpose
item_type_sub_item_map = {item_type_tag_sets: (tags_key, canonical_tag_name_key)}
item_id_prefix_to_type_dict = {environments_prefix: item_type_environments, tenants_prefix: item_type_tenants}


class Config:
    def __init__(self):
        self.octopus_endpoint = None
        self.octopus_name = None
        self.api_key = None
        self.user_name = None
        self.password = None
        self.overwrite = False
        self.no_stdout = False
        self.current_path = None
        self.get_config()

    def get_config(self):
        code_path = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(code_path, folder_configurations, file_configuration)
        config_dict = load_file(config_file)
        self.octopus_endpoint = config_dict.get(octopus_endpoint_key)
        self.octopus_name = config_dict.get(octopus_name_key)
        self.api_key = config_dict.get(api_key_key)
        self.user_name = config_dict.get(user_name_key)
        self.password = config_dict.get(password_key)
        self.overwrite = config_dict.get(overwrite_key)
        self.no_stdout = config_dict.get(no_stdout_key)
        self.current_path = os.getcwd()


config = Config()


def log_info_print(local_logger=None, msg=""):
    local_logger.info(msg)
    if not config.no_stdout:
        print(msg)


# TODO once the migration supports clone packages; we do not have to pre-process VersioningStrategy for cloning project
def prepare_project_versioning_strategy(project=None):
    logger.warning(f"prepare project {project.get(name_key)} for cloning by removing packages from "
                   f"{versioning_strategy_key}; TODO once the migration supports clone packages; we can stop doing it")
    project.get(versioning_strategy_key)[donor_package_key] = None
    project.get(versioning_strategy_key)[donor_package_step_id_key] = None


def get_list_ids_one_type(item_type=None, space_id=None):
    list_items = get_one_type_to_list(item_type=item_type, space_id=space_id)
    return [item.get(id_key) for item in list_items]


def verify_space(space_id_name=None):
    list_spaces = get_one_type_to_list(item_type=item_type_spaces)
    space = find_item(lst=list_spaces, key=id_key, value=space_id_name)
    if space:
        return space.get(id_key)
    space = find_single_item_from_list_by_name(list_items=list_spaces, item_name=space_id_name)
    if space:
        return space.get(id_key)
    return None


# remove the unnecessary modified date/user information from put and post operations
def pop_last_modified(a_dict=None):
    if isinstance(a_dict, dict):
        a_dict.pop('LastModifiedOn', None)
        a_dict.pop('LastModifiedBy', None)


def always_overwrite_or_compare_overwrite(local_file=None, data=None, overwrite=False):
    if not local_file or not str(data):
        raise ValueError("local_file and data must not be empty")
    if config.overwrite or overwrite:
        save_file(file_path_name=local_file, content=data)
        logger.info(f'A new local file {local_file} was written with the data')
    else:
        compare_overwrite(data=data, local_file=local_file)


# local child item file based on parent item
def get_local_child_file(parent_name=None, child_type=None, space_id=None):
    parent_name = parent_name.replace(slash_sign, underscore_sign)
    child_type = child_type.replace(slash_sign, underscore_sign)
    if space_id:
        return os.path.join(config.current_path, config.octopus_name, space_id, child_type,
                            parent_name + underscore_sign + child_type + yaml_ext)
    else:
        return os.path.join(config.current_path, config.octopus_name, folder_outer_spaces, child_type,
                            parent_name + underscore_sign + child_type + yaml_ext)


# get the local single item file from config.current_path, space_id, item type, file_name;
# for spaces files, no space_id needed
def get_local_single_item_file(item_name=None, item_type=None, space_id=None):
    item_type = item_type.replace(slash_sign, underscore_sign)
    item_name = item_name.replace(slash_sign, underscore_sign)
    if space_id:
        return os.path.join(config.current_path, config.octopus_name, space_id, item_type, item_name + yaml_ext)
    else:
        return os.path.join(config.current_path, config.octopus_name, folder_outer_spaces, item_type,
                            item_name + yaml_ext)


# get the local all items file from config.current_path, space_id, item type;
# for spaces files, no space_id needed
def get_local_all_items_file(item_type=None, space_id=None):
    item_type_name = item_type.replace(slash_sign, underscore_sign)
    if space_id:
        return os.path.join(config.current_path, config.octopus_name, space_id, item_type_name,
                            all_underscore + item_type_name + yaml_ext)
    else:
        return os.path.join(config.current_path, config.octopus_name, folder_outer_spaces, item_type_name,
                            all_underscore + item_type_name + yaml_ext)


# get local item file smartly; three possibilities
# use 'Name'
# use 'Id' if 'Name' not available
# use item_type if neither 'Id' nor 'Name' available
def get_local_single_item_file_from_item(item=None, item_type=None, space_id=None):
    if not item or not item_type:
        raise ValueError("item and item_type must not be empty!")
    if item.get(name_key) and item_type not in item_types_with_duplicate_names:
        local_item_file = get_local_single_item_file(item_name=item[name_key], item_type=item_type, space_id=space_id)
    elif item.get(id_key):
        local_item_file = get_local_single_item_file(item_name=item[id_key], item_type=item_type, space_id=space_id)
    else:
        local_item_file = get_local_single_item_file(item_name=item_type, item_type=item_type, space_id=space_id)
    return local_item_file


# check if the local item file is the same as the remote item on Octopus server;
# the remote item will be retrieved on the fly
def is_local_same_as_remote(item_type=None, item_name=None, item_id=None, space_id=None):
    if not item_type or not item_name and not item_id:
        raise ValueError("item_type and item_name/item_id must not be empty")
    if item_name:
        remote_item = get_single_item_by_name(item_type=item_type, item_name=item_name, space_id=space_id)
        local_item_file = get_local_single_item_file(item_name=item_name, item_type=item_type, space_id=space_id)
    else:
        remote_item = get_or_delete_single_item_by_id(item_type=item_type, item_id=item_id, space_id=space_id)
        local_item_file = get_local_single_item_file(item_name=item_id, item_type=item_type, space_id=space_id)
    is_same, local_item = is_local_same_as_remote2(remote_item=remote_item, local_item_file=local_item_file)
    return is_same, local_item, remote_item


# check if the local item file is the same as the item on Octopus server
# the remote item is an input
def is_local_same_as_remote2(remote_item=None, local_item_file=None):
    if not remote_item or not local_item_file:
        raise ValueError('remote_item and local_item_file must not be empty')
    local_item = load_file(local_item_file)
    return compare_dicts(local_item, remote_item), local_item


# compare in memory all items with the local all items and overwrite if user wants
def compare_overwrite_multiple_items(items=None, item_type=None, space_id=None, overwrite=False):
    if not item_type:
        raise ValueError('item_type must not be empty')
    local_all_items_file = get_local_all_items_file(item_type=item_type, space_id=space_id)
    logger.info('compare and write: ' + local_all_items_file)
    always_overwrite_or_compare_overwrite(local_file=local_all_items_file, data=items, overwrite=overwrite)


# get all items for an item_type by call Octopus API /api/{space_id}/item_type with 'get' operation
# {space_id} is optional
def get_one_type_ignore_error(item_type=None, space_id=None):
    logger.info(f"getting all items of {item_type} in space {space_id}")
    if not item_type:
        raise ValueError("item_type must not be empty")
    space_url = space_id + slash_sign if space_id else ""
    try:
        if item_type == item_type_home:
            logger.info(f"getting space {space_id} home page")
            return call_octopus(config=config, url_suffix=space_url)
        if item_type in only_all_types_inside_space:
            logger.info(f"{item_type} can only be downloaded by {slash_all}")
            url_suffix = space_url + item_type + slash_all
            return call_octopus(config=config, url_suffix=url_suffix)
        else:
            url_suffix = space_url + item_type + url_all_pages
            return call_octopus(config=config, url_suffix=url_suffix)
    except ValueError as err:
        # TODO bug https://help.octopus.com/t/504-gateway-time-out-on-getting-all-variables/24732
        logger.error(err)
        return {}


# get extended types like /api/users/{id}/permissions
def get_ext_types_save(item_type=None, space_id=None, item_ids=None):
    ext_types = ext_types_map.get(item_type)
    logger.info(f"Get extended types {ext_types} of {item_type} in Space {space_id}")
    for ext_type in ext_types:
        ext_items_dict = {}
        for item_id in item_ids:
            address = item_type + slash_sign + item_id + slash_sign + ext_type
            ext_item = request_octopus_item(payload=space_id, space_id=space_id, address=address)
            ext_items_dict[item_id] = ext_item
        ext_file = get_local_single_item_file(item_name=all_underscore + item_type + underscore_sign + ext_type,
                                              item_type=item_type, space_id=space_id)
        save_file(file_path_name=ext_file, content=ext_items_dict)


# then save the all items into a local file (warning for overwrite)
def get_one_type_save(item_type=None, space_id=None, overwrite=False):
    if not item_type:
        raise ValueError("item_type must not be empty")
    log_info_print(local_logger=logger, msg=f"downloading {item_type} in space {space_id}...")
    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
    compare_overwrite_multiple_items(items=all_items, item_type=item_type, space_id=space_id, overwrite=overwrite)
    if item_type == item_type_users:
        list_items = get_list_items_from_all_items(all_items=all_items)
        item_ids = [item.get(id_key) for item in list_items]
        get_ext_types_save(item_type=item_type_users, space_id=space_id, item_ids=item_ids)
    return all_items


# delete all items for an item_type by call Octopus API /api/{space_id}/item_type
# then save the all items into a local file (warning for overwrite)
def delete_one_type(item_type=None, space_id=None):
    if not item_type:
        raise ValueError("item_type must not be empty")
    log_info_print(local_logger=logger, msg=f"deleting all {item_type} in space {space_id}...")
    if not config.overwrite:
        if input(f"Delete all items of {item_type} in {space_id} [Y/n]? ") == 'Y':
            config.overwrite = True
        elif input(f"Delete NONE items of {item_type} in {space_id} [Y/n]? ") == 'Y':
            return
    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
    if item_type in item_types_without_single_item:
        logger.info(f"{item_type} has no sub-single-item, exit")
        return
    for item in get_list_items_from_all_items(all_items=all_items):
        delete_single_item_by_name_or_id(item_type=item_type, item_id=item.get(id_key), space_id=space_id)


# get all items for all item_type(s) by call Octopus API /api/{space_id}/item_type with 'get' operation
# item_types can be None, "", or "projects,tenants" etc
def get_types_save(item_types_comma_delimited=None, space_id=None):
    if item_types_comma_delimited:
        list_item_types = item_types_comma_delimited.split(comma_sign)
    else:
        if space_id:
            list_item_types = item_types_inside_space
        else:
            list_item_types = item_types_inside_space + item_types_only_ourter_space
    if config.overwrite:
        logger.info(f"===== You are downloading {list_item_types} from space {space_id}... ===== ")
    else:
        config.overwrite = \
            input(f"***** You are downloading {list_item_types} from space {space_id}; "
                  f"Some entities may already be downloaded locally; "
                  f"Do you want to overwrite all local existing entities? "
                  f"If no, you will be asked to overwrite or not for each type respectively. [Y/n]: ") == 'Y'
    for item_type in list_item_types:
        get_one_type_save(item_type=item_type, space_id=space_id)


def get_spaces_save(item_types_comma_delimited=None, space_id_or_name_comma_delimited=None):
    if space_id_or_name_comma_delimited:
        list_space_ids_or_names = space_id_or_name_comma_delimited.split(comma_sign)
        list_space_ids = [verify_space(space_id_name=space_id_or_name) for space_id_or_name in
                          list_space_ids_or_names]
    else:
        list_space_ids = get_list_ids_one_type(item_type=item_type_spaces) + [None]
    list_space_ids_set = set(list_space_ids)
    if config.overwrite:
        logger.info(f"===== You are downloading spaces {list_space_ids_set}... =====")
    else:
        config.overwrite = \
            input(f"===== You are downloading spaces {list_space_ids_set}; "
                  f"Some entities may already be downloaded locally; "
                  f"Do you want to overwrite all local existing entities? "
                  f"If no, you will be asked to overwrite or not for each type respectively. [Y/n]: ") == 'Y'
    for space_id in list_space_ids_set:
        get_types_save(item_types_comma_delimited=item_types_comma_delimited, space_id=space_id)


# get a single item from Octopus server
# 1. get all items for an item_type by call Octopus API /api/{space_id}/item_type with 'get' operation
# 2. find the matching item for the item_name
def get_single_item_by_name(item_type=None, item_name=None, space_id=None):
    if not item_type or not item_name:
        raise ValueError("item_type and item_name must not be empty")
    logger.info(f"Getting {item_type} {item_name} from {space_id} "
                f"by getting all items first and then find the matched item by name")
    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
    return find_single_item_from_list_by_name(list_items=all_items.get(items_key, []), item_name=item_name)


# find a single item by name from a list of items
def find_single_item_from_list_by_name(list_items=None, item_name=None):
    logger.info(f"Find {item_name} from list of items...")
    if not list_items:
        logger.info(f"The list is empty, so return")
        return {}
    item = find_item(list_items, name_key, item_name)
    if item.get(id_key):
        logger.info(f"{id_key} for {item_name} is " + item.get(id_key))
    else:
        logger.info(f"{item_name} has no {id_key}; the item is: ")
        logger.warning(pformat(item))
    return item


def save_single_item(item_type=None, item=None, space_id=None):
    if not item_type or not item:
        raise ValueError("item_type and item must not be empty")
    local_item_file = get_local_single_item_file_from_item(item=item, item_type=item_type, space_id=space_id)
    # always_overwrite_or_compare_overwrite(local_file=local_item_file, data=item)
    save_file(file_path_name=local_item_file, content=item)
    logger.info(f'A local file {local_item_file} was saved or overwritten with the data')
    return item


# get tenant variables
def get_tenant_variables(tenant_id=None, space_id=None):
    address = item_type_tenants + slash_sign + tenant_id + slash_sign + item_type_variables
    return request_octopus_item(space_id=space_id, address=address)


# get tenant variables and save to a file
def get_tenant_variables_save(tenant_id=None, space_id=None):
    logger.info(f"getting and saving {item_type_tenant_variables} for tenant {tenant_id} in space {space_id}...")
    tenant_variables = get_tenant_variables(tenant_id=tenant_id, space_id=space_id)
    dst_file = get_local_single_item_file(item_name=tenant_id + underscore_sign + item_type_variables,
                                          item_type=item_type_tenant_variables, space_id=space_id)
    save_file(file_path_name=dst_file, content=tenant_variables)
    return tenant_variables


# put/post tenant variables
def put_post_tenant_variables_save(tenant_id=None, space_id=None, tenant_variables=None):
    address = item_type_tenants + slash_sign + tenant_id + slash_sign + item_type_variables
    remote_tenant_variables = request_octopus_item(space_id=space_id, address=address)
    if remote_tenant_variables:
        logger.info(f"tenant {tenant_id} has existing variables, so put the variables")
        remote_tenant_variables = request_octopus_item(payload=tenant_variables, space_id=space_id, address=address,
                                                       action=operation_put)
    else:
        # TODO add a log to see if any "POST" exist, it may be an Octopus bug
        logger.info(f"tenant {tenant_id} has no variables, so post the variables")
        remote_tenant_variables = request_octopus_item(payload=tenant_variables, space_id=space_id, address=address,
                                                       action=operation_post)
    tenant_variables_file = get_local_single_item_file(item_name=tenant_id + underscore_sign + item_type_variables,
                                                       item_type=item_type_tenant_variables, space_id=space_id)
    save_file(file_path_name=tenant_variables_file, content=remote_tenant_variables)
    return remote_tenant_variables


def get_single_item_by_name_or_id(item_type=None, item_name=None, item_id=None, space_id=None):
    if not item_type or not item_name and not item_id:
        raise ValueError("item_type and item_name/item_id must not be empty")
    if item_name:
        return get_single_item_by_name(item_type=item_type, item_name=item_name, space_id=space_id)
    elif item_id:
        return get_or_delete_single_item_by_id(item_type=item_type, item_id=item_id, space_id=space_id)
    else:
        raise ValueError("Either item_name or item_id must be present")


# get a single item from Octopus server
# if item_name
# 1. get all items for an item_type by call Octopus API /api/{space_id}/item_type with 'get' operation
# 2. find the matching item for the item_name
# 3. save the single item into a local file (warning for overwrite)
# if item_id
# get a single item from Octopus server for the item which cannot be searched by the item_name (like deployment process)
# by directly calling the Octopus API /api/{space_id}/item_type/{id} with 'get'
# since there is no 'Name' in some of the json response, we have to use 'Id' as the file name to save it
def get_single_item_by_name_or_id_save(item_type=None, item_name=None, item_id=None, space_id=None):
    item_badge = item_name if item_name else item_id
    log_info_print(local_logger=logger,
                   msg=f"getting {item_type} {item_badge} in space {space_id} and saving to local file...")
    item = get_single_item_by_name_or_id(item_type=item_type, item_name=item_name, item_id=item_id, space_id=space_id)
    save_single_item(item_type=item_type, item=item, space_id=space_id)
    # process child items
    if item_type == item_type_projects:
        logger.info(f"the item type is {item_type_projects}, so also get its deployment_process and variables")
        get_single_item_by_name_or_id_save(item_type=item_type_deployment_processes,
                                           item_id=item.get(deployment_process_id_key), space_id=space_id)
        get_single_item_by_name_or_id_save(item_type=item_type_variables, item_id=item.get(variable_set_id_key),
                                           space_id=space_id)
    elif item_type == item_type_library_variable_sets:
        logger.info(f"the item type is {item_type_library_variable_sets}, so also get variables")
        get_single_item_by_name_or_id_save(item_type=item_type_variables, item_id=item.get(variable_set_id_key),
                                           space_id=space_id)
    elif item_type == item_type_tenants:
        logger.info(f"the item type is {item_type_tenants}, so also get its variables")
        get_tenant_variables_save(tenant_id=item.get(id_key), space_id=space_id)
    return item


# a single item from Octopus server for the item which cannot be searched by the item_name (like deployment process)
# by directly calling the Octopus API /api/{space_id}/item_type/{id}
def get_or_delete_single_item_by_id(item_type=None, item_id=None, action=operation_get, space_id=None):
    logger.info(f"{action} {item_type} {item_id} in {space_id}...")
    if not item_type or not item_id:
        raise ValueError("item_type and item_id must not be empty")
    space_url = space_id + slash_sign if space_id else ""
    url_suffix = space_url + item_type + slash_sign + item_id
    return call_octopus(operation=action, config=config, url_suffix=url_suffix)


def get_list_items_from_all_items(all_items=None):
    logger.info(f"getting the list of items from the payload")
    # the case where the payload has a metadata and a list
    if isinstance(all_items, dict) and isinstance(all_items.get(items_key), list):
        logger.info(f"the payload is a dict, so get the list of items first by {items_key}")
        return all_items.get(items_key)
    elif isinstance(all_items, list):
        logger.info("the payload is a list")
        return all_items
    return []


# post a single item by call Octopus API /api/{space_id}/item_type with 'post'
def post_single_item(item_type=None, payload=None, space_id=None):
    if not item_type or not payload:
        raise ValueError("item_type and playload must not be empty")
    space_url = space_id + slash_sign if space_id else ""
    url_suffix = space_url + item_type
    item = call_octopus(operation=operation_post, payload=payload, config=config, url_suffix=url_suffix)
    pop_last_modified(item)
    return item


# post a single item by call Octopus API /api/{space_id}/item_type with 'post' operation
# then save the item locally
def post_single_item_save(item_type=None, payload=None, space_id=None):
    item_badge = payload.get(name_key) if payload.get(name_key) else payload.get(id_key)
    logger.info(f"posting a new {item_type} {item_badge} in space {space_id} and saving it to a local file")
    item = post_single_item(item_type=item_type, payload=payload, space_id=space_id)
    if not item.get(name_key) and payload.get(name_key):
        logger.warning(f"the new item has no name, so the input item name {payload.get(name_key)} is used")
        item[name_key] = payload.get(name_key)
    local_item_file = get_local_single_item_file_from_item(item=item, item_type=item_type, space_id=space_id)
    always_overwrite_or_compare_overwrite(local_file=local_item_file, data=item, overwrite=True)
    return item


# put a single item by call Octopus API /api/{space_id}/item_type/{id} with 'put' operation
def put_single_item(item_type=None, payload=None, space_id=None):
    if not item_type or not payload:
        raise ValueError("item_type and playload must not be empty")
    logger.info(f"put a single {item_type} {payload.get(id_key)} to space {space_id}")
    space_url = space_id + slash_sign if space_id else ""
    # some type has no id like http://server/api/Spaces-1/dashboardconfiguration
    url_suffix = space_url + item_type + (slash_sign + payload.get(id_key) if payload.get(id_key) else "")
    item = call_octopus(operation=operation_put, payload=payload, config=config, url_suffix=url_suffix)
    pop_last_modified(item)
    logger.info(f"{item_type} {id_key} is " + item[id_key])
    return item


# put a single item by call Octopus API /api/{space_id}/item_type/{id} with 'put' operation
# then save the item locally
def put_single_item_save(item_type=None, payload=None, space_id=None, overwrite=False):
    item_info = payload.get(name_key) if payload.get(name_key) else payload.get(id_key)
    logger.info(f"updating {item_type} {item_info} in {space_id} and saving to a local file...")
    if config.overwrite or overwrite \
            or input(f"Are you sure you want to update {item_type} {item_info} in {space_id} [Y/n]: ") == 'Y':
        item = put_single_item(item_type=item_type, payload=payload, space_id=space_id)
        local_item_file = get_local_single_item_file_from_item(item=item, item_type=item_type, space_id=space_id)
        save_file(file_path_name=local_item_file, content=item)
        return item
    return payload


# put a child-item by call Octopus API /api/{space_id}/child_type/{id} with 'put' operation
# then save the item locally
def put_child_item_save(parent_name=None, child_type=None, payload=None, space_id=None):
    logger.info(f"put child {child_type} {payload.get(id_key)} of parent {parent_name} in space {space_id}")
    child_item = put_single_item(item_type=child_type, payload=payload, space_id=space_id)
    local_child_file = get_local_child_file(parent_name=parent_name, child_type=child_type, space_id=space_id)
    save_file(file_path_name=local_child_file, content=child_item)
    return child_item


# update an item on Octopus server
# 1. check if the local item file is the same as the remote item on Octopus server
# 2. if same, exit, otherwise continue to ask if user wants to update the item on Octopus server using the local item
# 3. if user input 'Y', update the item on Octopus and save the remote item locally
def update_single_item_save(item_type=None, item_name=None, item_id=None, space_id=None):
    if not item_type or not item_name and not item_id:
        raise ValueError("item_type and item_name/item_id must not be empty")
    log_info_print(local_logger=logger, msg=f"updating {item_type} {item_name if item_name else item_id} in space "
                                            f"{space_id} from local file")
    is_same, local_item, remote_item = \
        is_local_same_as_remote(item_type=item_type, item_name=item_name, item_id=item_id, space_id=space_id)
    if is_same:
        log_info_print(local_logger=logger, msg=remote_local_same_msg)
        return local_item
    return put_single_item_save(item_type=item_type, payload=local_item, space_id=space_id)


def delete_file(file_name=None):
    logger.info(f"Removing {file_name}")
    try:
        os.remove(file_name)
    except OSError:
        logger.info(f"{file_name} does not exit")
        pass


# delete unused sub items if the item cannot be deleted due to some sub items are being used
def delete_sub_items(item_type=None, item_name=None, item_id=None, space_id=None):
    item_badge = item_name if item_name else item_id
    sub_item_tuple = item_type_sub_item_map.get(item_type)
    if not sub_item_tuple:
        logger.warning(f"{item_type} {item_badge} does not have sub items for deleting; exit")
        return
    item = get_single_item_by_name_or_id(item_type=item_type, item_name=item_name, item_id=item_id, space_id=space_id)
    if not item:
        logger.warning(f"{item_type} {item_badge} does not exist in {space_id}; exit")
        return
    sub_items_key = sub_item_tuple[0]
    sub_item_name_key = sub_item_tuple[1]
    log_info_print(local_logger=logger,
                   msg=f"deleting the unused sub items {sub_items_key} of {item_type} {item_badge} in {space_id}...")
    index = 0
    while index < len(item.get(sub_items_key)):
        item_copy = copy.deepcopy(item)
        sub_item = item_copy.get(sub_items_key).pop(index)
        try:
            item = put_single_item(item_type=item_type, payload=item_copy, space_id=space_id)
            log_info_print(local_logger=logger,
                           msg=f"sub item with {sub_item_name_key}: {sub_item.get(sub_item_name_key)} was deleted")
        except ValueError as err:
            logger.warning(err)
            log_info_print(local_logger=logger, msg=f"sub item with {sub_item_name_key}: "
                                                    f"{sub_item.get(sub_item_name_key)} cannot be deleted; skip it")
            index += 1
    logger.info(f"----- completed deleting sub items -----")


# delete a single item
def delete_single_item_by_name_or_id(item_type=None, item_name=None, item_id=None, space_id=None):
    if not item_type or not item_name and not item_id:
        raise ValueError("item_type and item_name/item_id must not be empty")
    item_info = item_name if item_name else item_id
    item = get_single_item_by_name_or_id(item_type=item_type, item_name=item_name, item_id=item_id, space_id=space_id)
    if not item:
        logger.warning(f"{item_type} {item_info} does not exist in {space_id}; exit")
        return
    if not config.overwrite and input(f"Are you sure to delete {item_type} {item_info} in {space_id} [Y/n]: ") != 'Y':
        return
    log_info_print(local_logger=logger, msg=f"deleting {item_type} {item_info} in {space_id}...")
    try:
        get_or_delete_single_item_by_id(item_type=item_type, item_id=item.get(id_key), action=operation_delete,
                                        space_id=space_id)
    except ValueError as err:
        logger.warning(err)
        log_info_print(local_logger=logger, msg=f"{item_type} {item_info} cannot be deleted and try to delete unused "
                                                f"sub items in {space_id}...")
        delete_sub_items(item_type=item_type, item_id=item.get(id_key), space_id=space_id)


# create a new single item from a local file
def create_single_item_from_local_file(item_type=None, item_name=None, local_item_name=None, space_id=None):
    if not item_type or not local_item_name:
        raise ValueError('item_type and local_item_name must not be empty')
    log_info_print(local_logger=logger,
                   msg=f"creating a new {item_type} {item_name} in space {space_id} from a local file named "
                       f"{local_item_name}")
    # load local item file
    local_item_file = get_local_single_item_file(item_name=local_item_name, item_type=item_type, space_id=space_id)
    local_item = load_file(local_item_file)
    # rename item if needed
    if item_name:
        logger.info(f"Rename local item to {item_name}")
        local_item[name_key] = item_name
    remote = post_single_item_save(item_type=item_type, payload=local_item, space_id=space_id)
    return remote


# create a new single item from an item on Octopus server
def clone_single_item_from_remote_item(item_type=None, item_name=None, base_item_name=None, space_id=None):
    if not item_type or not item_name or not base_item_name:
        raise ValueError('item_type and item_name and base_item_name must not be empty')
    # read the remote base item
    log_info_print(local_logger=logger,
                   msg=f"cloning {item_type} {item_name} in space {space_id} based on remote item {base_item_name}")
    base_item = get_single_item_by_name(item_type=item_type, item_name=base_item_name, space_id=space_id)
    if item_type == item_type_projects:
        prepare_project_versioning_strategy(project=base_item)
    # create a new item from the remote item
    base_item[name_key] = item_name
    item = post_single_item_save(item_type=item_type, payload=base_item, space_id=space_id)
    return item


# get a single child-item of a parent item; the child-item has no 'Name' field, e.g. deployment processes of projects
def get_child_item(parent_name=None, parent_type=None, child_id_key=None, child_type=None, space_id=None):
    logger.info(f"getting child {child_type} of parent {parent_type} {parent_name} by {child_id_key} in {space_id}")
    parent = get_single_item_by_name(item_type=parent_type, item_name=parent_name, space_id=space_id)
    return get_or_delete_single_item_by_id(item_type=child_type, item_id=parent.get(child_id_key), space_id=space_id)


# get a single child-item of a parent item; the child-item has no 'Name' field
# save the child-item into a local file with the file name of parent_name_child_type
def get_child_item_save(parent_name=None, parent_type=None, child_id_key=None, child_type=None, space_id=None):
    log_info_print(local_logger=logger, msg=f"getting child {child_type} of parent {parent_type} {parent_name} by "
                                            f"{child_id_key} in {space_id}")
    child_item = get_child_item(parent_name=parent_name, parent_type=parent_type, child_id_key=child_id_key,
                                child_type=child_type, space_id=space_id)
    local_child_file = get_local_child_file(parent_name=parent_name, child_type=child_type, space_id=space_id)
    always_overwrite_or_compare_overwrite(local_file=local_child_file, data=child_item)
    return child_item


# update a single child-item of a parent item; the child-item has no 'Name' field
# 1. check if the local child-item file is the same as the remote child-item on Octopus server
# 2. if same, exit, otherwise ask if user wants to update the child-item on Octopus server using the local child-item
# 3. if user input 'Y', update the child-item on Octopus and save the remote child-item locally, otherwise exit
def update_child_item_from_local_save(
        parent_name=None, parent_type=None, child_id_key=None, child_type=None, space_id=None):
    log_info_print(local_logger=logger,
                   msg=f"update child {child_type} of {parent_type} {parent_name} in {space_id} by {child_id_key}")
    remote_child_item = get_child_item(parent_name=parent_name, parent_type=parent_type, child_id_key=child_id_key,
                                       child_type=child_type, space_id=space_id)
    local_child_file = get_local_child_file(parent_name=parent_name, child_type=child_type, space_id=space_id)
    is_same, local_child_item = is_local_same_as_remote2(remote_item=remote_child_item,
                                                         local_item_file=local_child_file)
    if is_same:
        log_info_print(local_logger=logger, msg=remote_local_same_msg)
        return
    if config.overwrite or input(f"Are you sure you want to update {child_type} for {parent_name} [Y/n]: ") == 'Y':
        child_item = put_single_item(item_type=child_type, payload=local_child_item, space_id=space_id)
        save_file(file_path_name=local_child_file, content=child_item)


# clone a child-item of a parent-item from another parent-item
def clone_child_item_from_another_parent_save(parent_name=None, base_parent_name=None, parent_type=None,
                                              child_id_key=None, child_type=None, sub_item_key=None, space_id=None,
                                              dst_space_id=None):
    if not dst_space_id:
        dst_space_id = space_id
    logger.info(f"Clone {parent_type} {base_parent_name}'s child item {child_type} in {space_id} "
                f"to {parent_type} {parent_name} in {dst_space_id}")
    base_child_item = get_child_item(parent_name=base_parent_name, parent_type=parent_type, child_id_key=child_id_key,
                                     child_type=child_type, space_id=space_id)
    return __clone_child_item_save(dst_space_id=dst_space_id, parent_name=parent_name,
                                   parent_type=parent_type, child_type=child_type, child_id_key=child_id_key,
                                   sub_item_key=sub_item_key, base_child_item=base_child_item,
                                   overwrite=config.overwrite)


def clone_item_by_id_replace_sub_item_save(item_type=None, src_item=None, dst_item_id=None, sub_item_key=None,
                                           dst_space_id=None):
    src_item_id = src_item.get(id_key)
    logger.info(f"By replacing {sub_item_key}, clone {item_type} {src_item_id} from memory to "
                f"{dst_item_id} in {dst_space_id}...")
    dst_item = get_or_delete_single_item_by_id(item_type=item_type, item_id=dst_item_id, space_id=dst_space_id)
    dst_item[sub_item_key] = src_item.get(sub_item_key)
    dst_item = put_single_item_save(item_type=item_type, payload=dst_item, space_id=dst_space_id, overwrite=True)
    return dst_item


def __clone_child_item_save(dst_space_id=None, parent_name=None, parent_type=None, child_type=None,
                            child_id_key=None, sub_item_key=None, base_child_item=None, overwrite=False):
    logger.info(f"Cloning child item {child_type} {base_child_item.get(id_key)} from memory to {parent_type} "
                f"{parent_name} in {dst_space_id}...")
    if config.overwrite or overwrite or input(f"Are you sure you want to clone child item {child_type} "
                                              f"{base_child_item.get(id_key)} from memory to {parent_type} "
                                              f"{parent_name} in {dst_space_id} [Y/n]: ") == 'Y':
        dst_parent = get_single_item_by_name(item_type=parent_type, item_name=parent_name, space_id=dst_space_id)
        dst_child_id = dst_parent.get(child_id_key)
        dst_child_item = clone_item_by_id_replace_sub_item_save(item_type=child_type, src_item=base_child_item,
                                                                dst_item_id=dst_child_id, sub_item_key=sub_item_key,
                                                                dst_space_id=dst_space_id)
        return dst_child_item


def merge_local_to_remote(source_item=None, target_item=None, child_id_key=None):
    source_children = source_item[child_id_key]
    target_children = target_item[child_id_key]
    target_children_names = set()
    for target_child in target_children:
        target_children_names.add(target_child[name_key])
    for source_child in source_children:
        if not source_child[name_key] in target_children_names:
            source_child[id_key] = ""
            target_children.append(source_child)


def merge_single_item_save(item_type=None, item_name=None, item_id=None, child_id_key=None, space_id=None):
    if not item_type or not item_name and not item_id or not child_id_key:
        raise ValueError("item_type and item_name/item_id and child_id_key must not be empty")
    log_info_print(local_logger=logger,
                   msg=f"merging {item_type} {item_name if item_name else item_id} for child {child_id_key} in space "
                       f"{space_id} from local file...")
    is_same, local_item, remote_item = is_local_same_as_remote(item_type=item_type, item_name=item_name,
                                                               item_id=item_id, space_id=space_id)
    if is_same:
        log_info_print(local_logger=logger, msg=remote_local_same_msg)
        return local_item
    merge_local_to_remote(source_item=local_item, target_item=remote_item, child_id_key=child_id_key)
    return put_single_item_save(item_type=item_type, payload=remote_item, space_id=space_id)


def get_list_items_from_file(item_type=None, space_id=None):
    if not item_type:
        raise ValueError("item_type must not be empty!")
    all_items_file = get_local_all_items_file(item_type=item_type, space_id=space_id)
    logger.info(f"load all_items file {all_items_file}...")
    all_items = load_file(all_items_file)
    return get_list_items_from_all_items(all_items=all_items)


def find_child_item_from_list(parent=None, list_items=None, child_id_key=None):
    item_id_value = parent.get(child_id_key)
    logger.info(f"Find {parent.get(name_key)}'s child item with ({child_id_key} = {item_id_value})")
    return find_item(lst=list_items, key=id_key, value=item_id_value)


# request an item by calling Octopus API /api/{space_id}/address
def request_octopus_item(payload=None, space_id=None, address=None, action=operation_get):
    space_url = space_id + slash_sign if space_id else ""
    url_suffix = space_url + address
    item = call_octopus(operation=action, payload=payload, config=config, url_suffix=url_suffix)
    pop_last_modified(item)
    return item


def find_sub_by_item(item_type=None, item_id=None, sub_type=None, sub_name=None, space_id=None):
    address = item_type + slash_sign + item_id + slash_sign + sub_type
    sub_items = request_octopus_item(payload=None, space_id=space_id, address=address, action=operation_get)
    list_sub_items = get_list_items_from_all_items(all_items=sub_items)
    return find_item(lst=list_sub_items, key=name_key, value=sub_name)


def get_item_id_by_name(item_type=None, item_name=None, space_id=None):
    item = get_single_item_by_name(item_type=item_type, item_name=item_name, space_id=space_id)
    return item.get(id_key)


def get_list_variables_by_set_name_or_id(set_name=None, set_id=None, space_id=None):
    if set_name:
        library_variable_set = \
            get_single_item_by_name(item_type=item_type_library_variable_sets, item_name=set_name, space_id=space_id)
        set_id = library_variable_set.get(variable_set_id_key)
    variables_dict = \
        get_or_delete_single_item_by_id(item_type=item_type_variables, item_id=set_id, space_id=space_id)
    return variables_dict.get(variables_key)


def get_one_type_to_list(item_type=None, space_id=None):
    all_items = get_one_type_ignore_error(item_type=item_type, space_id=space_id)
    return get_list_items_from_all_items(all_items=all_items)


def get_task_status(task_id=None, space_id=None):
    log_info_print(local_logger=logger, msg=f"check the status of task {task_id} in space {space_id}")
    task = get_or_delete_single_item_by_id(item_type=item_type_tasks, item_id=task_id, space_id=space_id)
    save_single_item(item_type=item_type_tasks, item=task, space_id=space_id)
    log_info_print(local_logger=logger,
                   msg=f"the task's status is {task.get(state_key)} and description is: {task.get(description_key)}")
    return task.get(state_key)


def wait_task(task_id=None, space_id=None, time_limit_second=600):
    log_info_print(local_logger=logger, msg=f"wait for task {task_id} in space {space_id} to complete until time out "
                                            f"at {time_limit_second} seconds")
    counter = 0
    while get_task_status(task_id=task_id, space_id=space_id) == executing_string:
        if counter > time_limit_second:
            log_info_print(local_logger=logger,
                           msg=f"task {task_id} takes longer than {time_limit_second} seconds and times out")
            return executing_string
        time.sleep(1)
        counter += 1
    status = get_task_status(task_id=task_id, space_id=space_id)
    log_info_print(local_logger=logger,
                   msg=f"task {task_id} in space {space_id} completes with status {status} at {counter} seconds")
    return status



import ast
import copy
import json
import logging
from pathlib import Path
from pprint import pformat

import yaml

yaml_exts = (".yaml", ".yml")
json_ext = ".json"

logger = logging.getLogger(__name__)


def find_index(lst=None, key=None, value=None):
    logger.info(f"searching the first occurrence of {key} = {value} in the list and return the index...")
    for i, dic in enumerate(lst):
        if dic.get(key) == value:
            logger.info(f"found the first occurrence at index {i}")
            return i
    logger.warning(f"no matched item found")
    return -1


def find_item(lst=None, key=None, value=None):
    logger.info(f"searching the first occurrence of {key} = {value} in the list and return the item...")
    index = find_index(lst, key, value)
    if index < 0:
        return {}
    return lst[index]


# TODO Octopus treats string case insensitive for item names
def find_matched_sub_list(lst=None, match_dict=None, ignore_case=False):
    logger.info(f"Try to find the matched sub-list for a list against {match_dict}...")
    list_copy = copy.deepcopy(lst)
    # we must do reversely to avoid unexpected result on deleting by index
    for index in range(len(list_copy) - 1, -1, -1):
        logger.info(f"working on the index {index} of the input list")
        dic = list_copy[index]
        for match_key, match_value in match_dict.items():
            if ignore_case and dic.get(match_key) != match_value and dic.get(match_key).lower() == match_value.lower():
                logger.warning(f"for key {match_key}, value of the {index}th item in the list is {dic.get(match_key)}, "
                               f"and value of the match dict is {match_value}. They are case insensitively same")
            elif not ignore_case and dic.get(match_key) != match_value \
                    or ignore_case and dic.get(match_key).lower() != match_value.lower():
                del list_copy[index]
                break
    if list_copy:
        logger.info(f"{len(list_copy)} matched items are found from the input list.")
    else:
        logger.info(f"no matched items are found from the input list.")
    return list_copy


def replace_list_new_value(lst=None, match_dict=None, replace_dict=None):
    logger.info(f"replace with {replace_dict} in list matching the input {match_dict}")
    for index, dic in enumerate(lst):
        logger.info(f"working on the index {index} of the input list")
        match = True
        for match_key, match_value in match_dict.items():
            if dic.get(match_key) != match_value:
                match = False
                break
        if match:
            logger.info(f"index at {index} is a match; replace with {replace_dict}")
            for replace_key, replace_value in replace_dict.items():
                dic[replace_key] = replace_value


def load_json_file(file_path_name=None):
    logger.info(f"loading {file_path_name} ...")
    if Path(file_path_name).is_file():
        with open(file_path_name) as f:
            return json.load(f)
    log_raise_value_error(err=f"File {file_path_name} does not exit")


def save_json_file(file_path_name=None, content=None):
    logger.info(f"writing {file_path_name} ...")
    make_dir(file_path_name=file_path_name)
    with open(file_path_name, 'w', newline='\n') as fp:
        json.dump(content, fp, indent=2)


def load_yaml_file(file_path_name=None):
    logger.info(f"loading {file_path_name} ...")
    if Path(file_path_name).is_file():
        with open(file_path_name, 'r') as stream:
            return yaml.safe_load(stream)
    log_raise_value_error(err=f"File {file_path_name} does not exit")


def save_yaml_file(file_path_name=None, content=None):
    logger.info(f"writing {file_path_name} ...")
    make_dir(file_path_name=file_path_name)
    with open(file_path_name, 'w', newline='\n') as outfile:
        yaml.safe_dump(content, outfile, default_flow_style=False, sort_keys=False)


def make_dir(file_path_name=None):
    if not file_path_name:
        log_raise_value_error(err=f"file path cannot be empty")
    p = Path(file_path_name).parent
    p.mkdir(parents=True, exist_ok=True)


# compare two lists, same returns True, otherwise returns False
def compare_lists(list1, list2):
    pairs = zip(list1, list2)
    diff = [(x, y) for x, y in pairs if x != y]
    logger.info('List difference: ' + str(diff))
    return not diff


# compare two dictionaries, same returns True, otherwise returns False
def compare_dicts(dict1, dict2):
    diff1 = {k: dict1[k] for k in dict1 if k not in dict2 or dict1[k] != dict2[k]}
    diff2 = {k: dict2[k] for k in dict2 if k not in dict1 or dict1[k] != dict2[k]}
    logger.info('Dict difference: ' + str([diff1, diff2]))
    return dict1 == dict2


def compare_overwrite(data=None, local_file=None):
    logger.info(f"comparing and save local file {local_file}...")
    if not Path(local_file).is_file():
        logger.info(f"{local_file} does not exist, so just write it")
        save_file(file_path_name=local_file, content=data)
        return
    logger.info(f"{local_file} already exists, so compare and write it")
    local = None
    if local_file.endswith(json_ext):
        local = load_json_file(local_file)
    elif local_file.endswith(yaml_exts):
        local = load_yaml_file(local_file)
    else:
        log_raise_value_error(err=f"The file to be compared must be either {yaml_exts} or {json_ext} file. "
                                  f"Your file {local_file} is not one of them")
    if isinstance(data, list):
        is_sync = compare_lists(local, data)
    elif isinstance(data, dict):
        is_sync = compare_dicts(local, data)
    else:
        is_sync = str(local) == str(data)
    if not is_sync:
        overwrite = input('The remote data is different from your local data. ' +
                          f'Do you want to overwrite your local file {local_file} [Y/n]: ')
        if overwrite == 'Y':
            logger.info(f'Writing your local file {local_file} with the remote data...')
            save_file(file_path_name=local_file, content=data)
        else:
            logger.info('No overwritten happens')


def load_file(file_path_name=None):
    if file_path_name.endswith(json_ext):
        return load_json_file(file_path_name=file_path_name)
    elif file_path_name.endswith(yaml_exts):
        return load_yaml_file(file_path_name=file_path_name)
    else:
        log_raise_value_error(err=f"The file to be loaded must be either {yaml_exts} or {json_ext} file. "
                                  f"Your file {file_path_name} is not one of them")


def save_file(file_path_name=None, content=None):
    logger.info(f"writing a new local file {file_path_name}...")
    if file_path_name.endswith(json_ext):
        save_json_file(file_path_name=file_path_name, content=content)
    elif file_path_name.endswith(yaml_exts):
        save_yaml_file(file_path_name=file_path_name, content=content)
    else:
        log_raise_value_error(err=f"The file to be saved must be either {yaml_exts} or {json_ext} file. Your file "
                                  f"{file_path_name} is not one of them")


def get_dict_from_str(string=None):
    logger.info(f"converting str {string} to dict...")
    python_structure = ast.literal_eval(string)
    if not isinstance(python_structure, dict):
        log_raise_value_error(item=python_structure, err="string is not a python dictionary")
    logger.info(python_structure)
    return python_structure


def log_raise_value_error(local_logger=logger, item=None, err=None):
    local_logger.error(err)
    if item:
        local_logger.info(pformat(item))
    raise ValueError(err)



from setuptools import setup, find_packages

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ["requests>=2.20.0", "PyYAML>=5.1"]

setup(
    name="octopus-python-client",
    version="1.3.0",
    author="Tony Li",
    author_email="tonybest@gmail.com",
    description="Python script to manage Octopus deploy servers through the Octopus Restful APIs",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/tableau/octopus-python-client",
    package_dir={'': 'src'},
    packages=find_packages('src'),
    install_requires=requirements,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'octopus_python_client = octopus_python_client.main:main'
        ]
    }
)
